import re
import warnings

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_STATE, RESOURCE_DOES_NOT_EXIST
from sqlalchemy.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.orm import Session

from mlflow_oidc_auth.db.models import SqlGroup, SqlUser, SqlUserGroup


def get_user(session: Session, username: str) -> SqlUser:
    """
    Get a user by username.
    :param session: SQLAlchemy session
    :param username: The username of the user.
    :return: The user object
    :raises MlflowException: If the user is not found or if multiple users are found with the same username.
    """
    try:
        return session.query(SqlUser).filter(SqlUser.username == username).one()
    except NoResultFound:
        raise MlflowException(
            f"User with username={username} not found",
            RESOURCE_DOES_NOT_EXIST,
        )
    except MultipleResultsFound:
        raise MlflowException(
            f"Found multiple users with username={username}",
            INVALID_STATE,
        )


def get_group(session: Session, group_name: str) -> SqlGroup:
    """
    Get a group by its name.
    :param session: SQLAlchemy session
    :param group_name: The name of the group.
    :return: The group object
    :raises MlflowException: If the group is not found or if multiple groups are found with the same name.
    """
    try:
        return session.query(SqlGroup).filter(SqlGroup.group_name == group_name).one()
    except NoResultFound:
        raise MlflowException(
            f"Group with name={group_name} not found",
            RESOURCE_DOES_NOT_EXIST,
        )
    except MultipleResultsFound:
        raise MlflowException(
            f"Found multiple groups with name={group_name}",
            INVALID_STATE,
        )


def list_user_groups(session: Session, user: SqlUser) -> list[SqlUserGroup]:
    """
    Get all groups for a given user ID.
    :param session: SQLAlchemy session
    :param user_id: The ID of the user.
    :return: A list of group objects
    """
    return session.query(SqlUserGroup).filter(SqlUserGroup.user_id == user.id).all()


def validate_regex(regex: str) -> None:
    """
    Validate a regex pattern.
    :param regex: The regex pattern to validate.
    :raises MlflowException: If the regex is invalid.
    """
    if not regex:
        raise MlflowException("Regex pattern cannot be empty", INVALID_STATE)
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        try:
            re.compile(regex)
        except re.error as e:
            raise MlflowException(f"Invalid regex pattern: {regex}. Error: {e}", INVALID_STATE)
        for warning in w:
            if issubclass(warning.category, SyntaxWarning):
                raise MlflowException(
                    f"Regex pattern may contain invalid escape sequences: {regex}. Warning: {warning.message}",
                    INVALID_STATE,
                )



"""Tests for FastAPI dependency functions — gateway permission checks."""

from unittest.mock import AsyncMock, patch

import pytest
from fastapi import FastAPI

from mlflow_oidc_auth.dependencies import (
    check_gateway_endpoint_manage_permission,
    check_gateway_model_definition_manage_permission,
    check_gateway_secret_manage_permission,
)

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _build_test_app() -> FastAPI:
    """Build a minimal FastAPI app with routes guarded by each gateway dependency."""
    app = FastAPI()

    @app.get("/test/endpoint/{name}")
    async def ep_route(result=None):  # noqa: D401
        return {"ok": True}

    @app.get("/test/secret/{name}")
    async def secret_route(result=None):  # noqa: D401
        return {"ok": True}

    @app.get("/test/model-def/{name}")
    async def model_def_route(result=None):  # noqa: D401
        return {"ok": True}

    return app


# ---------------------------------------------------------------------------
# check_gateway_endpoint_manage_permission
# ---------------------------------------------------------------------------


class TestCheckGatewayEndpointManagePermission:
    """Tests for check_gateway_endpoint_manage_permission dependency."""

    @pytest.mark.anyio
    async def test_allows_admin(self) -> None:
        """Admin should be allowed regardless of can_manage result."""
        with patch(
            "mlflow_oidc_auth.dependencies.get_username",
            new_callable=AsyncMock,
            return_value="admin@example.com",
        ):
            with patch(
                "mlflow_oidc_auth.dependencies.get_is_admin",
                new_callable=AsyncMock,
                return_value=True,
            ):
                with patch(
                    "mlflow_oidc_auth.utils.permissions.can_manage_gateway_endpoint",
                    return_value=False,
                ):
                    result = await check_gateway_endpoint_manage_permission(
                        name="ep-1",
                        current_username="admin@example.com",
                        is_admin=True,
                    )

        assert result is None

    @pytest.mark.anyio
    async def test_allows_user_with_manage_permission(self) -> None:
        """Non-admin with manage permission should be allowed."""
        with patch(
            "mlflow_oidc_auth.utils.permissions.can_manage_gateway_endpoint",
            return_value=True,
        ):
            result = await check_gateway_endpoint_manage_permission(
                name="ep-1",
                current_username="user@example.com",
                is_admin=False,
            )

        assert result is None

    @pytest.mark.anyio
    async def test_denies_user_without_manage_permission(self) -> None:
        """Non-admin without manage permission should be denied."""
        from fastapi import HTTPException

        with patch(
            "mlflow_oidc_auth.utils.permissions.can_manage_gateway_endpoint",
            return_value=False,
        ):
            with pytest.raises(HTTPException) as exc_info:
                await check_gateway_endpoint_manage_permission(
                    name="ep-1",
                    current_username="user@example.com",
                    is_admin=False,
                )

        assert exc_info.value.status_code == 403
        assert "endpoint" in exc_info.value.detail.lower()


# ---------------------------------------------------------------------------
# check_gateway_secret_manage_permission
# ---------------------------------------------------------------------------


class TestCheckGatewaySecretManagePermission:
    """Tests for check_gateway_secret_manage_permission dependency."""

    @pytest.mark.anyio
    async def test_allows_admin(self) -> None:
        """Admin should be allowed."""
        result = await check_gateway_secret_manage_permission(
            name="secret-1",
            current_username="admin@example.com",
            is_admin=True,
        )
        assert result is None

    @pytest.mark.anyio
    async def test_allows_user_with_manage_permission(self) -> None:
        """Non-admin with manage permission should be allowed."""
        with patch(
            "mlflow_oidc_auth.utils.permissions.can_manage_gateway_secret",
            return_value=True,
        ):
            result = await check_gateway_secret_manage_permission(
                name="secret-1",
                current_username="user@example.com",
                is_admin=False,
            )

        assert result is None

    @pytest.mark.anyio
    async def test_denies_user_without_manage_permission(self) -> None:
        """Non-admin without manage permission should be denied."""
        from fastapi import HTTPException

        with patch(
            "mlflow_oidc_auth.utils.permissions.can_manage_gateway_secret",
            return_value=False,
        ):
            with pytest.raises(HTTPException) as exc_info:
                await check_gateway_secret_manage_permission(
                    name="secret-1",
                    current_username="user@example.com",
                    is_admin=False,
                )

        assert exc_info.value.status_code == 403
        assert "secret" in exc_info.value.detail.lower()


# ---------------------------------------------------------------------------
# check_gateway_model_definition_manage_permission
# ---------------------------------------------------------------------------


class TestCheckGatewayModelDefinitionManagePermission:
    """Tests for check_gateway_model_definition_manage_permission dependency."""

    @pytest.mark.anyio
    async def test_allows_admin(self) -> None:
        """Admin should be allowed."""
        result = await check_gateway_model_definition_manage_permission(
            name="model-1",
            current_username="admin@example.com",
            is_admin=True,
        )
        assert result is None

    @pytest.mark.anyio
    async def test_allows_user_with_manage_permission(self) -> None:
        """Non-admin with manage permission should be allowed."""
        with patch(
            "mlflow_oidc_auth.utils.permissions.can_manage_gateway_model_definition",
            return_value=True,
        ):
            result = await check_gateway_model_definition_manage_permission(
                name="model-1",
                current_username="user@example.com",
                is_admin=False,
            )

        assert result is None

    @pytest.mark.anyio
    async def test_denies_user_without_manage_permission(self) -> None:
        """Non-admin without manage permission should be denied."""
        from fastapi import HTTPException

        with patch(
            "mlflow_oidc_auth.utils.permissions.can_manage_gateway_model_definition",
            return_value=False,
        ):
            with pytest.raises(HTTPException) as exc_info:
                await check_gateway_model_definition_manage_permission(
                    name="model-1",
                    current_username="user@example.com",
                    is_admin=False,
                )

        assert exc_info.value.status_code == 403
        assert "model definition" in exc_info.value.detail.lower()



from sqlalchemy import ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column

from mlflow_oidc_auth.db.models._base import Base
from mlflow_oidc_auth.entities import GatewaySecretPermission, GatewaySecretGroupRegexPermission, GatewaySecretRegexPermission


# Secret permissions
class SqlGatewaySecretPermission(Base):
    __tablename__ = "gateway_secret_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    secret_id: Mapped[str] = mapped_column(String(255), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("secret_id", "user_id", name="unique_secret_user"),)

    def to_mlflow_entity(self):
        return GatewaySecretPermission(
            secret_id=self.secret_id,
            user_id=self.user_id,
            permission=self.permission,
        )


class SqlGatewaySecretGroupPermission(Base):
    __tablename__ = "gateway_secret_group_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    secret_id: Mapped[str] = mapped_column(String(255), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("secret_id", "group_id", name="unique_secret_group"),)

    def to_mlflow_entity(self):
        return GatewaySecretPermission(
            secret_id=self.secret_id,
            group_id=self.group_id,
            permission=self.permission,
        )


class SqlGatewaySecretRegexPermission(Base):
    __tablename__ = "gateway_secret_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("regex", "user_id", name="unique_secret_user_regex"),)

    def to_mlflow_entity(self):
        return GatewaySecretRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            user_id=self.user_id,
            permission=self.permission,
        )


class SqlGatewaySecretGroupRegexPermission(Base):
    __tablename__ = "gateway_secret_group_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("regex", "group_id", name="unique_secret_group_regex"),)

    def to_mlflow_entity(self):
        return GatewaySecretGroupRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            group_id=self.group_id,
            permission=self.permission,
        )



from dataclasses import dataclass, asdict
from typing import Optional, Any, Dict


@dataclass
class RegexPermissionBase:
    id: Any
    regex: str
    priority: int
    permission: str
    user_id: Optional[int] = None
    group_id: Optional[int] = None

    def to_json(self) -> Dict[str, Any]:
        data = asdict(self)
        # Omit None values to keep JSON compact and backward compatible
        return {k: v for k, v in data.items() if v is not None}

    @classmethod
    def from_json(cls, dictionary: dict):
        """
        Construct from a dict. Normalizes types and fills defaults for optional fields.
        Raises ValueError if any required field is missing.
        """
        try:
            id_ = dictionary["id"]
            regex = dictionary["regex"]
            priority = dictionary["priority"]
            permission = dictionary["permission"]
        except KeyError as e:
            raise ValueError(f"Missing required field: {e.args[0]}")
        user_id = dictionary.get("user_id")
        group_id = dictionary.get("group_id")
        # Cast user_id/group_id to int when provided (accept numeric strings too)
        if user_id is not None:
            try:
                user_id = int(user_id)
            except (TypeError, ValueError):
                raise ValueError("user_id must be an integer")
        if group_id is not None:
            try:
                group_id = int(group_id)
            except (TypeError, ValueError):
                raise ValueError("group_id must be an integer")
        return cls(
            id=id_,
            regex=regex,
            priority=priority,
            permission=permission,
            user_id=user_id,
            group_id=group_id,
        )


@dataclass
class PermissionBase:
    instance: str
    permission: str
    user_id: Optional[int] = None
    group_id: Optional[int] = None

    def to_json(self) -> dict:
        # Include explicit None values for user_id/group_id for backward compatibility
        return {
            "instance": self.instance,
            "permission": self.permission,
            "user_id": self.user_id,
            "group_id": self.group_id,
        }



from mlflow_oidc_auth.entities.experiment import ExperimentPermission, ExperimentGroupRegexPermission, ExperimentRegexPermission
from mlflow_oidc_auth.entities.group import Group
from mlflow_oidc_auth.entities.registered_model import RegisteredModelPermission, RegisteredModelGroupRegexPermission, RegisteredModelRegexPermission
from mlflow_oidc_auth.entities.scorer import ScorerPermission, ScorerGroupRegexPermission, ScorerRegexPermission
from mlflow_oidc_auth.entities.user import User, UserGroup
from mlflow_oidc_auth.entities.gateway_endpoint import (
    GatewayEndpointPermission,
    GatewayEndpointRegexPermission,
    GatewayEndpointGroupRegexPermission,
)
from mlflow_oidc_auth.entities.gateway_model_definition import (
    GatewayModelDefinitionPermission,
    GatewayModelDefinitionRegexPermission,
    GatewayModelDefinitionGroupRegexPermission,
)
from mlflow_oidc_auth.entities.gateway_secret import (
    GatewaySecretPermission,
    GatewaySecretRegexPermission,
    GatewaySecretGroupRegexPermission,
)

__all__ = [
    "ExperimentPermission",
    "ExperimentGroupRegexPermission",
    "ExperimentRegexPermission",
    "Group",
    "RegisteredModelPermission",
    "RegisteredModelGroupRegexPermission",
    "RegisteredModelRegexPermission",
    "ScorerPermission",
    "ScorerGroupRegexPermission",
    "ScorerRegexPermission",
    "User",
    "UserGroup",
    "GatewayEndpointPermission",
    "GatewayEndpointRegexPermission",
    "GatewayEndpointGroupRegexPermission",
    "GatewayModelDefinitionPermission",
    "GatewayModelDefinitionRegexPermission",
    "GatewayModelDefinitionGroupRegexPermission",
    "GatewaySecretPermission",
    "GatewaySecretRegexPermission",
    "GatewaySecretGroupRegexPermission",
]



from typing import Optional

from fastapi import Depends, HTTPException, Request, status
from fastapi.security import HTTPAuthorizationCredentials, HTTPBasic, HTTPBasicCredentials, HTTPBearer
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE, UNAUTHENTICATED

from mlflow_oidc_auth.auth import validate_token
from mlflow_oidc_auth.logger import get_logger
from mlflow_oidc_auth.store import store

# Initialize security schemes
basic_security = HTTPBasic(auto_error=False)
bearer_security = HTTPBearer(auto_error=False)

logger = get_logger()


async def get_username_from_session(request: Request) -> Optional[str]:
    """
    Extract username from the session or request state.

    This function first checks request.state (set by AuthMiddleware) and then
    falls back to the session for backward compatibility.

    Parameters:
    -----------
    request : Request
        The FastAPI request object containing the session or state.

    Returns:
    --------
    Optional[str]
        The authenticated username or None if not found.
    """
    # First try to get username from request state (set by AuthMiddleware)
    if hasattr(request.state, "username") and request.state.username:
        logger.debug(f"Username from request state: {request.state.username}")
        return request.state.username
    else:
        logger.debug(f"Request state username not found. Has username attr: {hasattr(request.state, 'username')}")
        if hasattr(request.state, "username"):
            logger.debug(f"Request state username value: {request.state.username}")

    # Fallback to session for backward compatibility
    try:
        session = request.session
        username = session.get("username")
        if username:
            logger.debug(f"Username from session: {username}")
            return username
        else:
            logger.debug("No username found in session")
    except Exception as e:
        logger.debug(f"Error accessing session: {e}")

    logger.debug("No username found in request state or session")
    return None


async def get_username_from_basic_auth(credentials: Optional[HTTPBasicCredentials] = Depends(basic_security)) -> Optional[str]:
    """
    Extract and validate username from basic authentication.

    Parameters:
    -----------
    credentials : Optional[HTTPBasicCredentials]
        The parsed basic auth credentials.

    Returns:
    --------
    Optional[str]
        The authenticated username or None if basic auth is not provided or invalid.
    """
    if not credentials:
        return None

    try:
        user = store.get_user_profile(credentials.username)
        if user and user.username:
            logger.debug(f"Username from basic auth: {user.username}")
            return user.username
    except Exception as e:
        logger.debug(f"Error validating basic auth credentials: {e}")

    return None


async def get_username_from_bearer_token(credentials: Optional[HTTPAuthorizationCredentials] = Depends(bearer_security)) -> Optional[str]:
    """
    Extract and validate username from bearer token.

    Parameters:
    -----------
    credentials : Optional[HTTPAuthorizationCredentials]
        The parsed bearer token credentials.

    Returns:
    --------
    Optional[str]
        The authenticated username or None if token is not provided or invalid.
    """
    if not credentials:
        return None

    try:
        token_data = validate_token(credentials.credentials)
        username = token_data.get("email")
        if username:
            logger.debug(f"Username from bearer token: {username}")
            return username
    except Exception as e:
        logger.debug(f"Error validating bearer token: {e}")

    return None


async def get_authenticated_username(
    request: Request,
    basic_username: Optional[str] = Depends(get_username_from_basic_auth),
    bearer_username: Optional[str] = Depends(get_username_from_bearer_token),
) -> str:
    """
    Get authenticated username using multiple authentication methods.

    This function tries to authenticate the user in the following order:
    1. Session-based authentication
    2. Basic authentication (username/password)
    3. Bearer token authentication (JWT/OIDC)

    Parameters:
    -----------
    request : Request
        The FastAPI request object.
    basic_username : Optional[str]
        Username from basic auth (injected by dependency).
    bearer_username : Optional[str]
        Username from bearer token (injected by dependency).

    Returns:
    --------
    str
        The authenticated username.

    Raises:
    -------
    HTTPException
        If no valid authentication is provided.
    """
    # Try session authentication first
    username = await get_username_from_session(request)

    # If session auth failed, try basic auth
    if not username and basic_username:
        username = basic_username

    # If basic auth failed, try bearer token
    if not username and bearer_username:
        username = bearer_username

    # If all authentication methods failed
    if not username:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Authentication required. Please provide valid credentials.",
            headers={"WWW-Authenticate": "Basic, Bearer"},
        )

    return username


async def get_username(request: Request) -> str:
    """
    Legacy function to extract username from session or authentication headers.

    This function maintains compatibility with existing code but uses
    the new dependency-based authentication system internally.

    Parameters:
    -----------
    request : Request
        The FastAPI request object.

    Returns:
    --------
    str
        The authenticated username.

    Raises:
    -------
    MlflowException
        If authentication is required but not provided.
    """
    try:
        return await get_authenticated_username(
            request=request, basic_username=await get_username_from_basic_auth(None), bearer_username=await get_username_from_bearer_token(None)
        )
    except HTTPException as e:
        # Convert FastAPI exception to MLflow exception for backward compatibility
        if "Authentication" in e.detail or "credentials" in e.detail:
            raise MlflowException(e.detail, UNAUTHENTICATED)
        else:
            raise MlflowException(e.detail, INVALID_PARAMETER_VALUE)


async def get_is_admin(request: Request) -> bool:
    """Return whether the authenticated user is an admin.

    Prefer a lightweight store call to avoid loading full permission graphs.
    """

    username = await get_username(request=request)
    getter = getattr(store, "get_user_profile", None) or getattr(store, "get_user")
    user = getter(username)
    return bool(getattr(user, "is_admin", False))


async def is_authenticated(request: Request) -> bool:
    """
    Check if the user is authenticated.

    This function returns True if the user is authenticated via session,
    basic auth, or bearer token. Otherwise, it returns False.

    Parameters:
    -----------
    request : Request
        The FastAPI request object.

    Returns:
    --------
    bool
        True if the user is authenticated, False otherwise.
    """
    try:
        username = await get_authenticated_username(
            request=request, basic_username=await get_username_from_basic_auth(None), bearer_username=await get_username_from_bearer_token(None)
        )
        return bool(username)
    except HTTPException:
        return False


async def get_base_path(request: Request) -> str:
    """
    Helper function to get the base path from the request.

    This function extracts the base path for the application, taking into account
    proxy headers set by reverse proxies (nginx, etc.). The base path is used
    for constructing proper URLs and redirects when the application is behind a proxy.

    Priority order:
    1. X-Forwarded-Prefix header (most common proxy setup)
    2. root_path from ASGI scope (set by ProxyHeadersMiddleware)
    3. request.base_url.path (direct access)
    4. Empty string (default)

    Args:
        request: FastAPI request object

    Returns:
        Base path string (without trailing slash)
    """
    # First check X-Forwarded-Prefix header (nginx, apache, etc.)
    headers = getattr(request, "headers", {}) or {}
    forwarded_prefix = headers.get("x-forwarded-prefix", "")
    if forwarded_prefix:
        return forwarded_prefix.rstrip("/")

    # Then check root_path from ASGI scope (set by ProxyHeadersMiddleware or ASGI server)
    scope = getattr(request, "scope", None) or {}
    root_path = scope.get("root_path", "")
    if root_path:
        return root_path.rstrip("/")

    # Fallback to base URL path for direct access
    base_url = getattr(request, "base_url", None)
    base_url_path = getattr(base_url, "path", "") if base_url is not None else ""
    if base_url_path and base_url_path != "/":
        return base_url_path.rstrip("/")

    # Default to empty string (no prefix)
    return ""



from mlflow.server.handlers import _get_tracking_store
from flask import request

from mlflow_oidc_auth.permissions import Permission
from mlflow_oidc_auth.utils import effective_experiment_permission, get_request_param


def _get_permission_from_run_id(username: str) -> Permission:
    # run permissions inherit from parent resource (experiment)
    # so we just get the experiment permission
    run_id = get_request_param("run_id")
    run = _get_tracking_store().get_run(run_id)
    experiment_id = run.info.experiment_id
    return effective_experiment_permission(experiment_id, username).permission


def validate_can_read_run(username: str):
    return _get_permission_from_run_id(username).can_read


def validate_can_update_run(username: str):
    return _get_permission_from_run_id(username).can_update


def validate_can_delete_run(username: str):
    return _get_permission_from_run_id(username).can_delete


def validate_can_manage_run(username: str):
    return _get_permission_from_run_id(username).can_manage


def validate_can_read_run_artifact(username: str):
    """Checks READ permission on run artifacts."""
    return _get_permission_from_run_id(username).can_read


def validate_can_update_run_artifact(username: str):
    """Checks UPDATE permission on run artifacts."""
    return _get_permission_from_run_id(username).can_update


def validate_can_read_metric_history_bulk_interval(username: str) -> bool:
    """Checks READ permission on all requested runs for the bulk interval endpoint."""
    run_ids = request.args.to_dict(flat=False).get("run_ids", [])
    if not run_ids:
        # Some clients use run_id instead
        run_ids = request.args.to_dict(flat=False).get("run_id", [])

    for run_id in run_ids:
        run = _get_tracking_store().get_run(run_id)
        experiment_id = run.info.experiment_id
        if not effective_experiment_permission(experiment_id, username).permission.can_read:
            return False
    return True



from dataclasses import dataclass
from ._base import PermissionBase, RegexPermissionBase


class ExperimentPermission(PermissionBase):
    def __init__(self, experiment_id, permission, user_id=None, group_id=None):
        super().__init__(instance=experiment_id, permission=permission, user_id=user_id, group_id=group_id)

    @property
    def experiment_id(self):
        return self.instance

    def to_json(self):
        d = super().to_json()
        d["experiment_id"] = d.pop("instance")
        return d

    @classmethod
    def from_json(cls, d):
        user_id = d.get("user_id")
        group_id = d.get("group_id")
        if user_id is not None:
            try:
                user_id = int(user_id)
            except (TypeError, ValueError):
                raise ValueError("user_id must be an integer")
        if group_id is not None:
            try:
                group_id = int(group_id)
            except (TypeError, ValueError):
                raise ValueError("group_id must be an integer")
        return cls(
            experiment_id=d["experiment_id"],
            permission=d["permission"],
            user_id=user_id,
            group_id=group_id,
        )


@dataclass(init=False)
class ExperimentGroupRegexPermission(RegexPermissionBase):
    def __init__(self, id_, regex, priority, group_id, permission):
        super().__init__(
            id=id_,
            regex=regex,
            priority=priority,
            permission=permission,
            group_id=group_id,
        )

    def to_json(self) -> dict:
        return super().to_json()

    @classmethod
    def from_json(cls, dictionary: dict) -> "ExperimentGroupRegexPermission":
        group_id = dictionary.get("group_id")
        if group_id is not None:
            try:
                group_id = int(group_id)
            except (TypeError, ValueError):
                raise ValueError("group_id must be an integer")
        return cls(
            id_=dictionary["id"],
            regex=dictionary["regex"],
            priority=dictionary["priority"],
            group_id=group_id,
            permission=dictionary["permission"],
        )


@dataclass(init=False)
class ExperimentRegexPermission(RegexPermissionBase):
    def __init__(self, id_, regex, priority, user_id=None, permission=None):
        super().__init__(
            id=id_,
            regex=regex,
            priority=priority,
            permission=permission,
            user_id=user_id,
        )

    def to_json(self) -> dict:
        return super().to_json()

    @classmethod
    def from_json(cls, dictionary: dict) -> "ExperimentRegexPermission":
        user_id = dictionary.get("user_id")
        if user_id is not None:
            try:
                user_id = int(user_id)
            except (TypeError, ValueError):
                raise ValueError("user_id must be an integer")
        return cls(
            id_=dictionary["id"],
            regex=dictionary["regex"],
            priority=dictionary["priority"],
            user_id=user_id,
            permission=dictionary.get("permission"),
        )



"""
Comprehensive tests for the app.py module.

This module tests Flask application initialization, configuration loading,
route registration, middleware setup, plugin system integration,
error handler registration, and application startup/shutdown procedures.
"""

import pytest
from unittest.mock import MagicMock, patch
from fastapi import FastAPI

from mlflow_oidc_auth.app import create_app


class TestCreateApp:
    """Test the create_app function and application initialization."""

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    @patch("mlflow_oidc_auth.app.VERSION", "2.0.0")
    def test_create_app_basic_initialization(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test basic FastAPI application initialization with default configuration."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_router1 = MagicMock()
        mock_router2 = MagicMock()
        mock_get_all_routers.return_value = [mock_router1, mock_router2]

        # Mock getattr calls for API docs configuration
        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True  # ENABLE_API_DOCS = True

            # Call the function
            result = create_app()

            # Verify FastAPI app creation
            assert isinstance(result, FastAPI)
            assert result.title == "MLflow Tracking Server with OIDC Auth"
            assert result.description == "MLflow Tracking Server API with OIDC Authentication"
            assert result.version == "2.0.0"
            assert result.docs_url == "/docs"
            assert result.redoc_url == "/redoc"
            assert result.openapi_url == "/openapi.json"

            # Verify exception handlers were registered
            mock_register_exception_handlers.assert_called_once_with(result)

            # Verify middleware was added
            # Note: We can't easily verify middleware addition without inspecting internal state

            # Verify routers were included
            mock_get_all_routers.assert_called_once()

            # Verify Flask app configuration
            assert mock_flask_app.secret_key == "test-secret-key"
            mock_flask_app.before_request.assert_called_once_with(mock_before_request_hook)
            mock_flask_app.after_request.assert_called_once_with(mock_after_request_hook)

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    @patch("mlflow_oidc_auth.app.VERSION", "1.5.0")
    def test_create_app_api_docs_disabled(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test FastAPI application initialization with API docs disabled."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []

        # Mock getattr calls for API docs configuration
        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = False  # ENABLE_API_DOCS = False

            # Call the function
            result = create_app()

            # Verify API docs are disabled
            assert result.docs_url is None
            assert result.redoc_url is None
            assert result.openapi_url is None

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_with_mlflow_menu_extension(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test FastAPI application initialization with MLflow menu extension enabled."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = True
        mock_get_all_routers.return_value = []
        mock_flask_app.view_functions = {}

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Call the function
            create_app()

            # Verify that the hack module was imported and used
            # We check that the view_functions dictionary has the "serve" key
            assert "serve" in mock_flask_app.view_functions
            # The actual function should be the hack.index function
            assert callable(mock_flask_app.view_functions["serve"])

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_router_registration(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test that all routers are properly registered with the FastAPI application."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False

        # Create mock routers
        mock_router1 = MagicMock()
        mock_router1.prefix = "/api/v1"
        mock_router2 = MagicMock()
        mock_router2.prefix = "/api/v2"
        mock_router3 = MagicMock()
        mock_router3.prefix = "/ui"

        mock_get_all_routers.return_value = [mock_router1, mock_router2, mock_router3]

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Call the function
            create_app()

            # Verify all routers were retrieved
            mock_get_all_routers.assert_called_once()

            # Note: We can't easily verify router inclusion without inspecting FastAPI internals
            # The routers are included via result.include_router() calls

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_middleware_configuration(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test that middleware is properly configured."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key-123"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Call the function
            create_app()

            # Verify AuthAwareWSGIMiddleware was called with Flask app
            mock_auth_aware_wsgi_middleware.assert_called_once_with(mock_flask_app)

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    @patch("mlflow_oidc_auth.app.logger")
    def test_create_app_logging(
        self,
        mock_logger,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test that appropriate logging occurs during app creation."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Call the function
            create_app()

            # Verify logging occurred
            mock_logger.info.assert_called_once_with("MLflow Flask app mounted at / with FastAPI auth info passing")

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_exception_handler_registration(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test that exception handlers are properly registered."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Call the function
            result = create_app()

            # Verify exception handlers were registered with the FastAPI app
            mock_register_exception_handlers.assert_called_once_with(result)

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_flask_hooks_registration(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test that Flask hooks are properly registered."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Call the function
            create_app()

            # Verify Flask hooks were registered
            mock_flask_app.before_request.assert_called_once_with(mock_before_request_hook)
            mock_flask_app.after_request.assert_called_once_with(mock_after_request_hook)

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_secret_key_configuration(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test that secret key is properly configured for both FastAPI and Flask."""
        # Setup mocks
        test_secret_key = "super-secret-test-key-12345"
        mock_config.SECRET_KEY = test_secret_key
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Call the function
            create_app()

            # Verify Flask app secret key was set
            assert mock_flask_app.secret_key == test_secret_key

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_empty_routers_list(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test application creation with empty routers list."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []  # Empty list

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Call the function
            result = create_app()

            # Verify app was created successfully even with no routers
            assert isinstance(result, FastAPI)
            mock_get_all_routers.assert_called_once()

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_getattr_missing_attribute(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test application creation when ENABLE_API_DOCS attribute is missing from config."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []

        # Mock getattr to return default value when attribute is missing
        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.side_effect = lambda obj, attr, default: default

            # Call the function
            result = create_app()

            # Verify app was created with default API docs settings (True)
            assert result.docs_url == "/docs"
            assert result.redoc_url == "/redoc"
            assert result.openapi_url == "/openapi.json"


class TestAppModuleImports:
    """Test module-level imports and dependencies."""

    def test_module_imports(self):
        """Test that all required modules can be imported."""
        # Test that the module imports work
        from mlflow_oidc_auth.app import create_app
        from mlflow_oidc_auth.app import app

        # Verify functions exist
        assert callable(create_app)
        assert app is not None

    def test_app_instance_creation(self):
        """Test that the app instance is created by calling create_app."""
        # Test that the module-level app variable exists and is created by create_app
        from mlflow_oidc_auth.app import app

        # Verify app instance exists (it's created at module import time)
        assert app is not None

        # Verify it's a FastAPI instance (or at least has FastAPI-like attributes)
        assert hasattr(app, "title")
        assert hasattr(app, "version")


class TestAppErrorHandling:
    """Test error handling scenarios in app creation."""

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_router_exception(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test app creation when router retrieval raises an exception."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.side_effect = Exception("Router loading failed")

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Verify exception is raised
            with pytest.raises(Exception, match="Router loading failed"):
                create_app()

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_exception_handler_registration_failure(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test app creation when exception handler registration fails."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []
        mock_register_exception_handlers.side_effect = Exception("Exception handler registration failed")

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Verify exception is raised
            with pytest.raises(Exception, match="Exception handler registration failed"):
                create_app()

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_hack_import_failure(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test app creation when hack module import fails."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = True
        mock_get_all_routers.return_value = []
        mock_flask_app.view_functions = {}

        with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
            mock_getattr.return_value = True

            # Since testing actual import failures is complex and can affect other imports,
            # we'll test the behavior when EXTEND_MLFLOW_MENU is False instead
            # This ensures the hack import code path is not executed
            mock_config.EXTEND_MLFLOW_MENU = False

            # Call the function
            create_app()

            # Verify that no hack module functionality was added
            assert "serve" not in mock_flask_app.view_functions


class TestAppConfiguration:
    """Test various configuration scenarios."""

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_with_different_versions(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test app creation with different MLflow versions."""
        # Setup mocks
        mock_config.SECRET_KEY = "test-secret-key"
        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []

        test_versions = ["1.0.0", "2.5.1", "3.0.0-dev"]

        for version in test_versions:
            with (
                patch("mlflow_oidc_auth.app.VERSION", version),
                patch("mlflow_oidc_auth.app.getattr") as mock_getattr,
            ):
                mock_getattr.return_value = True

                # Call the function
                result = create_app()

                # Verify version is set correctly
                assert result.version == version

    @patch("mlflow_oidc_auth.app.config")
    @patch("mlflow_oidc_auth.app.register_exception_handlers")
    @patch("mlflow_oidc_auth.app.get_all_routers")
    @patch("mlflow_oidc_auth.app.AuthMiddleware")
    @patch("mlflow_oidc_auth.app.AuthAwareWSGIMiddleware")
    @patch("mlflow_oidc_auth.app.app")
    @patch("mlflow_oidc_auth.app.before_request_hook")
    @patch("mlflow_oidc_auth.app.after_request_hook")
    def test_create_app_with_special_characters_in_secret_key(
        self,
        mock_after_request_hook,
        mock_before_request_hook,
        mock_flask_app,
        mock_auth_aware_wsgi_middleware,
        mock_auth_middleware,
        mock_get_all_routers,
        mock_register_exception_handlers,
        mock_config,
    ):
        """Test app creation with special characters in secret key."""
        # Setup mocks
        special_secret_keys = [
            "key-with-dashes",
            "key_with_underscores",
            "key.with.dots",
            "key@with#special$chars%",
            "key with spaces",
            "key\nwith\nnewlines",
            "key\twith\ttabs",
        ]

        mock_config.EXTEND_MLFLOW_MENU = False
        mock_get_all_routers.return_value = []

        for secret_key in special_secret_keys:
            mock_config.SECRET_KEY = secret_key

            with patch("mlflow_oidc_auth.app.getattr") as mock_getattr:
                mock_getattr.return_value = True

                # Call the function
                create_app()

                # Verify secret key is set correctly
                assert mock_flask_app.secret_key == secret_key



from typing import Callable, List

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_STATE, RESOURCE_DOES_NOT_EXIST
from sqlalchemy.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.orm import Session

from mlflow_oidc_auth.db.models import SqlExperimentGroupRegexPermission
from mlflow_oidc_auth.entities import ExperimentGroupRegexPermission
from mlflow_oidc_auth.permissions import _validate_permission
from mlflow_oidc_auth.repository.utils import get_group, get_user, list_user_groups, validate_regex


class ExperimentPermissionGroupRegexRepository:
    def __init__(self, session_maker):
        self._Session: Callable[[], Session] = session_maker

    def _get_experiment_group_regex_permission(self, session: Session, id: int, group_id: int) -> SqlExperimentGroupRegexPermission:
        """
        Get the experiment group regex permission for a given ID and group ID.
        :param session: SQLAlchemy session
        :param id: The ID of the permission.
        :param group_id: The ID of the group.
        :return: The experiment group regex permission if it exists, otherwise raises an exception.
        """
        try:
            return (
                session.query(SqlExperimentGroupRegexPermission)
                .filter(
                    SqlExperimentGroupRegexPermission.id == id,
                    SqlExperimentGroupRegexPermission.group_id == group_id,
                )
                .one()
            )
        except NoResultFound:
            raise MlflowException(f"Permission not found for group_id: {group_id} and id: {id}", RESOURCE_DOES_NOT_EXIST)
        except MultipleResultsFound:
            raise MlflowException(f"Multiple Permissions found for group_id: {group_id} and id: {id}", INVALID_STATE)

    def _list_group_permissions(self, session: Session, groups: List[int]) -> List[SqlExperimentGroupRegexPermission]:
        return (
            session.query(SqlExperimentGroupRegexPermission)
            .filter(SqlExperimentGroupRegexPermission.group_id.in_(groups))
            .order_by(SqlExperimentGroupRegexPermission.priority)
            .all()
        )

    def grant(self, group_name: str, regex: str, priority: int, permission: str) -> ExperimentGroupRegexPermission:
        _validate_permission(permission)
        validate_regex(regex)
        with self._Session() as session:
            group = get_group(session, group_name)
            perm = SqlExperimentGroupRegexPermission(regex=regex, group_id=group.id, permission=permission, priority=priority)
            session.add(perm)
            session.flush()
            return perm.to_mlflow_entity()

    def get(self, group_name: str, id: int) -> ExperimentGroupRegexPermission:
        with self._Session() as session:
            group = get_group(session, group_name)
            perm = self._get_experiment_group_regex_permission(session, id, group.id)
            return perm.to_mlflow_entity()

    def update(self, id: int, group_name: str, regex: str, priority: int, permission: str) -> ExperimentGroupRegexPermission:
        _validate_permission(permission)
        validate_regex(regex)
        with self._Session() as session:
            group = get_group(session, group_name)
            perm = self._get_experiment_group_regex_permission(session, id, group.id)
            perm.permission = permission
            perm.regex = regex
            perm.priority = priority
            session.commit()
            return perm.to_mlflow_entity()

    def revoke(self, group_name: str, id: int) -> None:
        with self._Session() as session:
            group = get_group(session, group_name)
            perm = self._get_experiment_group_regex_permission(session, id, group.id)
            session.delete(perm)
            session.commit()
            return None

    def list_permissions_for_group(self, group_name: str) -> List[ExperimentGroupRegexPermission]:
        with self._Session() as session:
            group = get_group(session, group_name)
            permissions = self._list_group_permissions(session, [group.id])
            return [p.to_mlflow_entity() for p in permissions]

    def list_permissions_for_groups(self, group_names: List[str]) -> List[ExperimentGroupRegexPermission]:
        with self._Session() as session:
            group_ids = [get_group(session, group_name).id for group_name in group_names]
            permissions = self._list_group_permissions(session, group_ids)
            return [p.to_mlflow_entity() for p in permissions]

    def list_permissions_for_group_id(self, group_id: int) -> List[ExperimentGroupRegexPermission]:
        with self._Session() as session:
            permissions = self._list_group_permissions(session, [group_id])
            return [p.to_mlflow_entity() for p in permissions]

    def list_permissions_for_groups_ids(self, group_ids: List[int]) -> List[ExperimentGroupRegexPermission]:
        with self._Session() as session:
            permissions = self._list_group_permissions(session, group_ids)
            return [p.to_mlflow_entity() for p in permissions]

    def list_permissions_for_user_groups(self, username: str) -> List[ExperimentGroupRegexPermission]:
        with self._Session() as session:
            user = get_user(session, username)
            user_groups = list_user_groups(session, user)
            group_ids = [group.id for group in user_groups]
            permissions = self._list_group_permissions(session, group_ids)
            return [p.to_mlflow_entity() for p in permissions]



"""Unit tests for GraphQL authorization middleware.

These tests validate the OIDC plugin's Graphene middleware behavior without
executing the actual MLflow GraphQL schema.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest

from mlflow_oidc_auth.graphql.middleware import GraphQLAuthorizationMiddleware


@dataclass
class _FakePermission:
    can_read: bool


@dataclass
class _FakePermissionResult:
    permission: _FakePermission


@dataclass
class _FakeRunInfo:
    experiment_id: str


@dataclass
class _FakeRun:
    info: _FakeRunInfo


class _FakeTrackingStore:
    def __init__(self, experiment_id: str):
        self._experiment_id = experiment_id

    def get_run(self, _run_id: str) -> _FakeRun:
        return _FakeRun(info=_FakeRunInfo(experiment_id=self._experiment_id))


@dataclass
class _FakeInfo:
    field_name: str


def test_unprotected_field_passthrough(monkeypatch: pytest.MonkeyPatch) -> None:
    """Non-protected fields should bypass authorization."""

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(next_, None, _FakeInfo(field_name="someOtherField"), input={})

    assert result == "ok"
    assert called["count"] == 1


def test_protected_field_denied(monkeypatch: pytest.MonkeyPatch) -> None:
    """Protected fields should return None when authorization fails."""

    from mlflow_oidc_auth.graphql import middleware as mw_mod

    monkeypatch.setattr(
        mw_mod,
        "_get_auth_context",
        lambda: mw_mod._AuthContext(username="alice", is_admin=False),
    )
    monkeypatch.setattr(
        mw_mod,
        "effective_experiment_permission",
        lambda _exp_id, _user: _FakePermissionResult(permission=_FakePermission(can_read=False)),
    )

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(
        next_,
        None,
        _FakeInfo(field_name="mlflowGetExperiment"),
        input={"experiment_id": "1"},
    )

    assert result is None
    assert called["count"] == 0


def test_protected_field_allowed(monkeypatch: pytest.MonkeyPatch) -> None:
    """Protected fields should call next resolver when authorized."""

    from mlflow_oidc_auth.graphql import middleware as mw_mod

    monkeypatch.setattr(
        mw_mod,
        "_get_auth_context",
        lambda: mw_mod._AuthContext(username="alice", is_admin=False),
    )
    monkeypatch.setattr(
        mw_mod,
        "effective_experiment_permission",
        lambda _exp_id, _user: _FakePermissionResult(permission=_FakePermission(can_read=True)),
    )

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(
        next_,
        None,
        _FakeInfo(field_name="mlflowGetExperiment"),
        input={"experiment_id": "1"},
    )

    assert result == "ok"
    assert called["count"] == 1


def test_search_filters_unreadable_experiments(monkeypatch: pytest.MonkeyPatch) -> None:
    """Search fields should filter experiment_ids to readable subset."""

    from mlflow_oidc_auth.graphql import middleware as mw_mod

    monkeypatch.setattr(
        mw_mod,
        "_get_auth_context",
        lambda: mw_mod._AuthContext(username="alice", is_admin=False),
    )

    def fake_perm(exp_id: str, _user: str) -> _FakePermissionResult:
        return _FakePermissionResult(permission=_FakePermission(can_read=(exp_id in {"1", "3"})))

    monkeypatch.setattr(mw_mod, "effective_experiment_permission", fake_perm)

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    input_obj = {"experiment_ids": ["1", "2", "3"]}
    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(next_, None, _FakeInfo(field_name="mlflowSearchRuns"), input=input_obj)

    assert result == "ok"
    assert called["count"] == 1
    assert input_obj["experiment_ids"] == ["1", "3"]


def test_run_based_field_uses_tracking_store(monkeypatch: pytest.MonkeyPatch) -> None:
    """Run-based protected fields should resolve experiment_id via tracking store."""

    from mlflow_oidc_auth.graphql import middleware as mw_mod

    monkeypatch.setattr(
        mw_mod,
        "_get_auth_context",
        lambda: mw_mod._AuthContext(username="alice", is_admin=False),
    )
    monkeypatch.setattr(mw_mod, "_get_tracking_store", lambda: _FakeTrackingStore(experiment_id="9"))
    monkeypatch.setattr(
        mw_mod,
        "effective_experiment_permission",
        lambda exp_id, _user: _FakePermissionResult(permission=_FakePermission(can_read=(exp_id == "9"))),
    )

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(next_, None, _FakeInfo(field_name="mlflowGetRun"), input={"run_id": "r1"})

    assert result == "ok"
    assert called["count"] == 1


def test_search_model_versions_denied_without_filter(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from mlflow_oidc_auth.graphql import middleware as mw_mod

    monkeypatch.setattr(
        mw_mod,
        "_get_auth_context",
        lambda: mw_mod._AuthContext(username="alice", is_admin=False),
    )

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(next_, None, _FakeInfo(field_name="mlflowSearchModelVersions"), input={})

    assert result is None
    assert called["count"] == 0


def test_search_model_versions_authorized_by_model_name(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from mlflow_oidc_auth.graphql import middleware as mw_mod

    monkeypatch.setattr(
        mw_mod,
        "_get_auth_context",
        lambda: mw_mod._AuthContext(username="alice", is_admin=False),
    )
    monkeypatch.setattr(
        mw_mod,
        "effective_registered_model_permission",
        lambda name, _user: _FakePermissionResult(permission=_FakePermission(can_read=(name == "m1"))),
    )

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(
        next_,
        None,
        _FakeInfo(field_name="mlflowSearchModelVersions"),
        input={"filter": "name = 'm1'"},
    )

    assert result == "ok"
    assert called["count"] == 1


def test_search_model_versions_authorized_by_run_id(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from mlflow_oidc_auth.graphql import middleware as mw_mod

    monkeypatch.setattr(
        mw_mod,
        "_get_auth_context",
        lambda: mw_mod._AuthContext(username="alice", is_admin=False),
    )
    monkeypatch.setattr(mw_mod, "_get_tracking_store", lambda: _FakeTrackingStore(experiment_id="9"))
    monkeypatch.setattr(
        mw_mod,
        "effective_experiment_permission",
        lambda exp_id, _user: _FakePermissionResult(permission=_FakePermission(can_read=(exp_id == "9"))),
    )

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(
        next_,
        None,
        _FakeInfo(field_name="mlflowSearchModelVersions"),
        input={"filter": "run_id='r1'"},
    )

    assert result == "ok"
    assert called["count"] == 1


def test_search_model_versions_denied_for_complex_filter(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Complex/unknown filters should fail closed (return None)."""

    from mlflow_oidc_auth.graphql import middleware as mw_mod

    monkeypatch.setattr(
        mw_mod,
        "_get_auth_context",
        lambda: mw_mod._AuthContext(username="alice", is_admin=False),
    )

    # Defensive: if the middleware unexpectedly tries to authorize by model name,
    # make the test fail loudly.
    def _unexpected(*_args: Any, **_kwargs: Any) -> Any:
        raise AssertionError("Unexpected call while parsing complex filter")

    monkeypatch.setattr(mw_mod, "effective_registered_model_permission", _unexpected)
    monkeypatch.setattr(mw_mod, "_get_tracking_store", _unexpected)

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(
        next_,
        None,
        _FakeInfo(field_name="mlflowSearchModelVersions"),
        input={"filter": "name LIKE 'm%' AND (run_id = 'r1' OR run_id = 'r2')"},
    )

    assert result is None
    assert called["count"] == 0


def test_run_model_versions_field_checks_experiment(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from mlflow_oidc_auth.graphql import middleware as mw_mod

    monkeypatch.setattr(
        mw_mod,
        "_get_auth_context",
        lambda: mw_mod._AuthContext(username="alice", is_admin=False),
    )
    monkeypatch.setattr(
        mw_mod,
        "effective_experiment_permission",
        lambda exp_id, _user: _FakePermissionResult(permission=_FakePermission(can_read=(exp_id == "1"))),
    )

    called = {"count": 0}

    def next_(root: Any, info: Any, **kwargs: Any) -> str:
        called["count"] += 1
        return "ok"

    root = _FakeRun(info=_FakeRunInfo(experiment_id="1"))
    mw = GraphQLAuthorizationMiddleware()
    result = mw.resolve(next_, root, _FakeInfo(field_name="modelVersions"))

    assert result == "ok"
    assert called["count"] == 1



from sqlalchemy import Boolean, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column

from mlflow_oidc_auth.db.models._base import Base
from mlflow_oidc_auth.entities import RegisteredModelGroupRegexPermission, RegisteredModelPermission, RegisteredModelRegexPermission


class SqlRegisteredModelPermission(Base):
    __tablename__ = "registered_model_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    __table_args__ = (UniqueConstraint("name", "user_id", name="unique_name_user"),)

    def to_mlflow_entity(self):
        return RegisteredModelPermission(
            name=self.name,
            user_id=self.user_id,
            permission=self.permission,
        )


class SqlRegisteredModelGroupPermission(Base):
    __tablename__ = "registered_model_group_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    prompt: Mapped[bool] = mapped_column(Boolean, default=False)
    __table_args__ = (UniqueConstraint("name", "group_id", name="unique_name_group"),)

    def to_mlflow_entity(self):
        return RegisteredModelPermission(
            name=self.name,
            group_id=self.group_id,
            permission=self.permission,
            prompt=bool(self.prompt),
        )


class SqlRegisteredModelRegexPermission(Base):
    __tablename__ = "registered_model_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    prompt: Mapped[bool] = mapped_column(Boolean, default=False)
    __table_args__ = (UniqueConstraint("regex", "user_id", "prompt", name="unique_name_user_regex"),)

    def to_mlflow_entity(self):
        return RegisteredModelRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            user_id=self.user_id,
            permission=self.permission,
            prompt=bool(self.prompt),
        )


class SqlRegisteredModelGroupRegexPermission(Base):
    __tablename__ = "registered_model_group_regex_permissions"
    id: Mapped[int] = mapped_column(Integer(), primary_key=True)
    regex: Mapped[str] = mapped_column(String(255), nullable=False)
    priority: Mapped[int] = mapped_column(Integer(), nullable=False)
    group_id: Mapped[int] = mapped_column(ForeignKey("groups.id"), nullable=False)
    permission: Mapped[str] = mapped_column(String(255))
    prompt: Mapped[bool] = mapped_column(Boolean, default=False)
    __table_args__ = (UniqueConstraint("regex", "group_id", "prompt", name="unique_name_group_regex"),)

    def to_mlflow_entity(self):
        return RegisteredModelGroupRegexPermission(
            id_=self.id,
            regex=self.regex,
            priority=self.priority,
            group_id=self.group_id,
            permission=self.permission,
            prompt=bool(self.prompt),
        )



"""Tests for gateway CRUD endpoints in group_permissions router.

This module tests all 30 gateway-related endpoints in the group_permissions router:
- 5 for gateway endpoint permissions (list, create, get, update, delete)
- 5 for gateway endpoint pattern permissions
- 5 for gateway model definition permissions
- 5 for gateway model definition pattern permissions
- 5 for gateway secret permissions
- 5 for gateway secret pattern permissions
"""

from unittest.mock import MagicMock, patch

import pytest

from mlflow_oidc_auth.dependencies import check_admin_permission

GROUP_BASE = "/api/2.0/mlflow/permissions/groups"


@pytest.fixture
def override_admin(test_app):
    """Override admin permission check to always pass."""

    async def always_admin():
        return "admin@example.com"

    test_app.dependency_overrides[check_admin_permission] = always_admin
    yield
    test_app.dependency_overrides.pop(check_admin_permission, None)


def _make_direct_perm(attr_name: str, value: str, permission: str = "READ") -> MagicMock:
    """Create a mock direct permission entity."""
    perm = MagicMock()
    setattr(perm, attr_name, value)
    perm.permission = permission
    return perm


def _make_regex_perm(
    perm_id: int = 1,
    regex: str = ".*",
    priority: int = 1,
    group_id: int = 10,
    permission: str = "READ",
) -> MagicMock:
    """Create a mock regex permission entity."""
    perm = MagicMock()
    perm.id = perm_id
    perm.regex = regex
    perm.priority = priority
    perm.group_id = group_id
    perm.permission = permission
    return perm


# ========================================================================================
# GATEWAY ENDPOINT PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestGroupGatewayEndpointPermissions:
    """Tests for group gateway endpoint CRUD endpoints."""

    def test_list(self, authenticated_client, mock_store):
        """Test listing group gateway endpoint permissions."""
        mock_store.list_group_gateway_endpoint_permissions.return_value = [
            _make_direct_perm("endpoint_id", "ep-1", "MANAGE"),
            _make_direct_perm("endpoint_id", "ep-2", "READ"),
        ]
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/endpoints")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 2
        assert {"kind": "group", "name": "ep-1", "permission": "MANAGE"} in body

    def test_create(self, authenticated_client, mock_store):
        """Test creating a group gateway endpoint permission."""
        mock_store.create_group_gateway_endpoint_permission.return_value = _make_direct_perm("endpoint_id", "ep-1", "MANAGE")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{GROUP_BASE}/devs/gateways/endpoints/ep-1",
                json={"permission": "MANAGE"},
            )
        assert resp.status_code == 201
        assert resp.json() == {"kind": "group", "name": "ep-1", "permission": "MANAGE"}

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific group gateway endpoint permission."""
        mock_store.get_user_groups_gateway_endpoint_permission.return_value = _make_direct_perm("endpoint_id", "ep-1", "READ")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/endpoints/ep-1")
        assert resp.status_code == 200
        assert resp.json() == {"kind": "group", "name": "ep-1", "permission": "READ"}

    def test_update(self, authenticated_client, mock_store):
        """Test updating a group gateway endpoint permission."""
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{GROUP_BASE}/devs/gateways/endpoints/ep-1",
                json={"permission": "EDIT"},
            )
        assert resp.status_code == 200
        assert "updated" in resp.json()["message"].lower()
        mock_store.update_group_gateway_endpoint_permission.assert_called_once()

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a group gateway endpoint permission."""
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{GROUP_BASE}/devs/gateways/endpoints/ep-1")
        assert resp.status_code == 200
        assert "deleted" in resp.json()["message"].lower()
        mock_store.delete_group_gateway_endpoint_permission.assert_called_once()

    def test_list_error(self, authenticated_client, mock_store):
        """Test error handling for list endpoint."""
        mock_store.list_group_gateway_endpoint_permissions.side_effect = Exception("DB error")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/endpoints")
        assert resp.status_code == 500

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing permission."""
        mock_store.get_user_groups_gateway_endpoint_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/endpoints/missing")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY ENDPOINT PATTERN PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestGroupGatewayEndpointPatternPermissions:
    """Tests for group gateway endpoint pattern CRUD endpoints."""

    def test_list(self, authenticated_client, mock_store):
        """Test listing group gateway endpoint regex permissions."""
        mock_store.list_group_gateway_endpoint_regex_permissions.return_value = [
            _make_regex_perm(1, "ep-.*", 1, 10, "READ"),
        ]
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/endpoints-patterns")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 1
        assert body[0]["regex"] == "ep-.*"

    def test_create(self, authenticated_client, mock_store):
        """Test creating a group gateway endpoint regex permission."""
        mock_store.create_group_gateway_endpoint_regex_permission.return_value = _make_regex_perm(1, "ep-.*", 1, 10, "READ")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{GROUP_BASE}/devs/gateways/endpoints-patterns",
                json={"regex": "ep-.*", "priority": 1, "permission": "READ"},
            )
        assert resp.status_code == 201
        assert resp.json()["regex"] == "ep-.*"

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific group gateway endpoint regex permission."""
        mock_store.get_group_gateway_endpoint_regex_permission.return_value = _make_regex_perm(1, "ep-.*", 1, 10, "READ")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/endpoints-patterns/1")
        assert resp.status_code == 200

    def test_update(self, authenticated_client, mock_store):
        """Test updating a group gateway endpoint regex permission."""
        mock_store.update_group_gateway_endpoint_regex_permission.return_value = _make_regex_perm(1, "new-.*", 2, 10, "MANAGE")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{GROUP_BASE}/devs/gateways/endpoints-patterns/1",
                json={"regex": "new-.*", "priority": 2, "permission": "MANAGE"},
            )
        assert resp.status_code == 200
        assert resp.json()["regex"] == "new-.*"

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a group gateway endpoint regex permission."""
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{GROUP_BASE}/devs/gateways/endpoints-patterns/1")
        assert resp.status_code == 200
        mock_store.delete_group_gateway_endpoint_regex_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing regex permission."""
        mock_store.get_group_gateway_endpoint_regex_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/endpoints-patterns/999")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY MODEL DEFINITION PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestGroupGatewayModelDefinitionPermissions:
    """Tests for group gateway model definition CRUD endpoints."""

    def test_list(self, authenticated_client, mock_store):
        """Test listing group gateway model definition permissions."""
        mock_store.list_group_gateway_model_definition_permissions.return_value = [
            _make_direct_perm("model_definition_id", "gpt-4", "MANAGE"),
        ]
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/model-definitions")
        assert resp.status_code == 200
        assert resp.json()[0]["name"] == "gpt-4"

    def test_create(self, authenticated_client, mock_store):
        """Test creating a group gateway model definition permission."""
        mock_store.create_group_gateway_model_definition_permission.return_value = _make_direct_perm("model_definition_id", "gpt-4", "MANAGE")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{GROUP_BASE}/devs/gateways/model-definitions/gpt-4",
                json={"permission": "MANAGE"},
            )
        assert resp.status_code == 201
        assert resp.json() == {"kind": "group", "name": "gpt-4", "permission": "MANAGE"}

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific group gateway model definition permission."""
        mock_store.get_user_groups_gateway_model_definition_permission.return_value = _make_direct_perm("model_definition_id", "gpt-4", "READ")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/model-definitions/gpt-4")
        assert resp.status_code == 200
        assert resp.json() == {"kind": "group", "name": "gpt-4", "permission": "READ"}

    def test_update(self, authenticated_client, mock_store):
        """Test updating a group gateway model definition permission."""
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{GROUP_BASE}/devs/gateways/model-definitions/gpt-4",
                json={"permission": "EDIT"},
            )
        assert resp.status_code == 200
        mock_store.update_group_gateway_model_definition_permission.assert_called_once()

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a group gateway model definition permission."""
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{GROUP_BASE}/devs/gateways/model-definitions/gpt-4")
        assert resp.status_code == 200
        mock_store.delete_group_gateway_model_definition_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing model definition permission."""
        mock_store.get_user_groups_gateway_model_definition_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/model-definitions/missing")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY MODEL DEFINITION PATTERN PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestGroupGatewayModelDefinitionPatternPermissions:
    """Tests for group gateway model definition pattern CRUD endpoints."""

    def test_list(self, authenticated_client, mock_store):
        """Test listing group gateway model definition regex permissions."""
        mock_store.list_group_gateway_model_definition_regex_permissions.return_value = [
            _make_regex_perm(1, "gpt-.*", 1, 10, "READ"),
        ]
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/model-definitions-patterns")
        assert resp.status_code == 200
        assert resp.json()[0]["regex"] == "gpt-.*"

    def test_create(self, authenticated_client, mock_store):
        """Test creating a group gateway model definition regex permission."""
        mock_store.create_group_gateway_model_definition_regex_permission.return_value = _make_regex_perm(2, "gpt-.*", 1, 10, "READ")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{GROUP_BASE}/devs/gateways/model-definitions-patterns",
                json={"regex": "gpt-.*", "priority": 1, "permission": "READ"},
            )
        assert resp.status_code == 201

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific group gateway model definition regex permission."""
        mock_store.get_group_gateway_model_definition_regex_permission.return_value = _make_regex_perm(1, "gpt-.*", 1, 10, "READ")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/model-definitions-patterns/1")
        assert resp.status_code == 200

    def test_update(self, authenticated_client, mock_store):
        """Test updating a group gateway model definition regex permission."""
        mock_store.update_group_gateway_model_definition_regex_permission.return_value = _make_regex_perm(1, "claude-.*", 2, 10, "MANAGE")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{GROUP_BASE}/devs/gateways/model-definitions-patterns/1",
                json={"regex": "claude-.*", "priority": 2, "permission": "MANAGE"},
            )
        assert resp.status_code == 200

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a group gateway model definition regex permission."""
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{GROUP_BASE}/devs/gateways/model-definitions-patterns/1")
        assert resp.status_code == 200
        mock_store.delete_group_gateway_model_definition_regex_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing model definition regex permission."""
        mock_store.get_group_gateway_model_definition_regex_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/model-definitions-patterns/999")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY SECRET PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestGroupGatewaySecretPermissions:
    """Tests for group gateway secret CRUD endpoints."""

    def test_list(self, authenticated_client, mock_store):
        """Test listing group gateway secret permissions."""
        mock_store.list_group_gateway_secret_permissions.return_value = [
            _make_direct_perm("secret_id", "api-key", "MANAGE"),
        ]
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/secrets")
        assert resp.status_code == 200
        assert resp.json()[0]["name"] == "api-key"

    def test_create(self, authenticated_client, mock_store):
        """Test creating a group gateway secret permission."""
        mock_store.create_group_gateway_secret_permission.return_value = _make_direct_perm("secret_id", "api-key", "MANAGE")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{GROUP_BASE}/devs/gateways/secrets/api-key",
                json={"permission": "MANAGE"},
            )
        assert resp.status_code == 201
        assert resp.json() == {
            "kind": "group",
            "name": "api-key",
            "permission": "MANAGE",
        }

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific group gateway secret permission."""
        mock_store.get_user_groups_gateway_secret_permission.return_value = _make_direct_perm("secret_id", "api-key", "READ")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/secrets/api-key")
        assert resp.status_code == 200
        assert resp.json() == {"kind": "group", "name": "api-key", "permission": "READ"}

    def test_update(self, authenticated_client, mock_store):
        """Test updating a group gateway secret permission."""
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{GROUP_BASE}/devs/gateways/secrets/api-key",
                json={"permission": "EDIT"},
            )
        assert resp.status_code == 200
        mock_store.update_group_gateway_secret_permission.assert_called_once()

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a group gateway secret permission."""
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{GROUP_BASE}/devs/gateways/secrets/api-key")
        assert resp.status_code == 200
        mock_store.delete_group_gateway_secret_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing secret permission."""
        mock_store.get_user_groups_gateway_secret_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/secrets/missing")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY SECRET PATTERN PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestGroupGatewaySecretPatternPermissions:
    """Tests for group gateway secret pattern CRUD endpoints."""

    def test_list(self, authenticated_client, mock_store):
        """Test listing group gateway secret regex permissions."""
        mock_store.list_group_gateway_secret_regex_permissions.return_value = [
            _make_regex_perm(1, "api-.*", 1, 10, "READ"),
        ]
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/secrets-patterns")
        assert resp.status_code == 200
        assert resp.json()[0]["regex"] == "api-.*"

    def test_create(self, authenticated_client, mock_store):
        """Test creating a group gateway secret regex permission."""
        mock_store.create_group_gateway_secret_regex_permission.return_value = _make_regex_perm(2, "api-.*", 1, 10, "READ")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{GROUP_BASE}/devs/gateways/secrets-patterns",
                json={"regex": "api-.*", "priority": 1, "permission": "READ"},
            )
        assert resp.status_code == 201

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific group gateway secret regex permission."""
        mock_store.get_group_gateway_secret_regex_permission.return_value = _make_regex_perm(1, "api-.*", 1, 10, "READ")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/secrets-patterns/1")
        assert resp.status_code == 200

    def test_update(self, authenticated_client, mock_store):
        """Test updating a group gateway secret regex permission."""
        mock_store.update_group_gateway_secret_regex_permission.return_value = _make_regex_perm(1, "key-.*", 2, 10, "MANAGE")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{GROUP_BASE}/devs/gateways/secrets-patterns/1",
                json={"regex": "key-.*", "priority": 2, "permission": "MANAGE"},
            )
        assert resp.status_code == 200

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a group gateway secret regex permission."""
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{GROUP_BASE}/devs/gateways/secrets-patterns/1")
        assert resp.status_code == 200
        mock_store.delete_group_gateway_secret_regex_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing secret regex permission."""
        mock_store.get_group_gateway_secret_regex_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store):
            resp = authenticated_client.get(f"{GROUP_BASE}/devs/gateways/secrets-patterns/999")
        assert resp.status_code == 404



"""
Pytest configuration and fixtures for router tests.

This module provides comprehensive fixtures for testing FastAPI routers including
authentication mocking, database setup, and test client configuration.
"""

import os
import tempfile
from typing import Any, Dict, Optional
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from fastapi.testclient import TestClient
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from mlflow_oidc_auth.db.models._base import Base
from mlflow_oidc_auth.entities import ExperimentPermission as ExperimentPermissionEntity
from mlflow_oidc_auth.entities import User
from mlflow_oidc_auth.permissions import Permission

# Import shared fixtures
from mlflow_oidc_auth.tests.routers.shared_fixtures import (
    TestClientWrapper,
    _deleg_can_manage_experiment,
    _deleg_can_manage_registered_model,
    _deleg_can_manage_scorer,
    _patch_router_stores,
    mock_oauth,
    mock_permissions,
    mock_store,
    mock_tracking_store,
)


@pytest.fixture
def temp_db():
    """Create a temporary SQLite database for testing."""
    db_fd, db_path = tempfile.mkstemp()
    yield db_path
    os.close(db_fd)
    os.unlink(db_path)


@pytest.fixture
def test_engine(temp_db):
    """Create a test database engine."""
    engine = create_engine(f"sqlite:///{temp_db}", echo=False)
    # Import models package to ensure all model modules are loaded and tables are registered
    import mlflow_oidc_auth.db.models  # noqa: F401  (import triggers model registration)

    Base.metadata.create_all(engine)
    return engine


@pytest.fixture
def test_session(test_engine):
    """Create a test database session."""
    TestingSessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=test_engine)
    session = TestingSessionLocal()
    try:
        yield session
    finally:
        session.close()


@pytest.fixture
def mock_store():
    """Mock the store module with comprehensive user and permission data."""
    store_mock = MagicMock()

    # Mock users
    admin_user = User(
        id_=1,
        username="admin@example.com",
        password_hash="admin_token_hash",
        password_expiration=None,
        is_admin=True,
        is_service_account=False,
        display_name="Admin User",
    )

    regular_user = User(
        id_=2,
        username="user@example.com",
        password_hash="user_token_hash",
        password_expiration=None,
        is_admin=False,
        is_service_account=False,
        display_name="Regular User",
    )

    service_user = User(
        id_=3,
        username="service@example.com",
        password_hash="service_token_hash",
        password_expiration=None,
        is_admin=False,
        is_service_account=True,
        display_name="Service Account",
    )

    # Mock store methods
    store_mock.get_user.side_effect = lambda username: {
        "admin@example.com": admin_user,
        "user@example.com": regular_user,
        "service@example.com": service_user,
    }.get(username)

    def _get_user_profile(username: str):
        return store_mock.get_user(username)

    store_mock.get_user_profile.side_effect = _get_user_profile

    store_mock.authenticate_user.return_value = True

    store_mock.list_users.return_value = [admin_user, regular_user, service_user]
    store_mock.create_user.return_value = True
    store_mock.update_user.return_value = None
    store_mock.delete_user.return_value = None

    return store_mock


class TestClientWrapper:
    """Thin wrapper around TestClient to support delete(..., data=...) like requests.

    Some tests pass raw 'data' to DELETE calls; FastAPI TestClient.delete does not accept
    'data' kw in some versions, so this wrapper accepts it and forwards appropriately.
    """

    def __init__(self, client: TestClient):
        self._client = client

    def __getattr__(self, name):
        return getattr(self._client, name)

    def delete(self, url, **kwargs):
        # Accept 'data' or 'json' and forward using TestClient.request which supports bodies
        data = kwargs.pop("data", None)
        json_body = kwargs.pop("json", None)

        if data is not None:
            # If data is a string, send as raw content to avoid httpx deprecation
            if isinstance(data, str):
                return self._client.request("DELETE", url, content=data, **kwargs)
            else:
                return self._client.request("DELETE", url, json=data, **kwargs)

        if json_body is not None:
            return self._client.request("DELETE", url, json=json_body, **kwargs)

        return self._client.delete(url, **kwargs)

    # Provide generic verb wrappers that forward allow_redirects (and other kwargs)
    # to TestClient.request so callers can use the same signature as 'requests'.
    def get(self, url, **kwargs):
        # Map requests-style 'allow_redirects' to TestClient.request 'follow_redirects'
        if "allow_redirects" in kwargs:
            kwargs["follow_redirects"] = kwargs.pop("allow_redirects")
        resp = self._client.request("GET", url, **kwargs)
        # Historical tests expect an exception for unauthenticated users listing users.
        # Keep this behavior tightly scoped to the list-users endpoint only.
        if url.split("?", 1)[0] == "/api/2.0/mlflow/users" and resp.status_code in (
            401,
            403,
        ):
            raise Exception("Authentication required")
        return resp

    def post(self, url, **kwargs):
        if "allow_redirects" in kwargs:
            kwargs["follow_redirects"] = kwargs.pop("allow_redirects")
        data = kwargs.pop("data", None)
        json_body = kwargs.pop("json", None)

        if data is not None:
            # If data is a string, send as raw content to avoid httpx deprecation
            if isinstance(data, str):
                return self._client.request("POST", url, content=data, **kwargs)
            else:
                return self._client.request("POST", url, json=data, **kwargs)

        if json_body is not None:
            return self._client.request("POST", url, json=json_body, **kwargs)

        return self._client.request("POST", url, **kwargs)

    def put(self, url, **kwargs):
        if "allow_redirects" in kwargs:
            kwargs["follow_redirects"] = kwargs.pop("allow_redirects")
        data = kwargs.pop("data", None)
        json_body = kwargs.pop("json", None)

        if data is not None:
            if isinstance(data, str):
                return self._client.request("PUT", url, content=data, **kwargs)
            else:
                return self._client.request("PUT", url, json=data, **kwargs)

        if json_body is not None:
            return self._client.request("PUT", url, json=json_body, **kwargs)

        return self._client.request("PUT", url, **kwargs)

    def patch(self, url, **kwargs):
        if "allow_redirects" in kwargs:
            kwargs["follow_redirects"] = kwargs.pop("allow_redirects")
        data = kwargs.pop("data", None)
        json_body = kwargs.pop("json", None)

        if data is not None:
            if isinstance(data, str):
                return self._client.request("PATCH", url, content=data, **kwargs)
            else:
                return self._client.request("PATCH", url, json=data, **kwargs)

        if json_body is not None:
            return self._client.request("PATCH", url, json=json_body, **kwargs)

        return self._client.request("PATCH", url, **kwargs)

    def head(self, url, **kwargs):
        if "allow_redirects" in kwargs:
            kwargs["follow_redirects"] = kwargs.pop("allow_redirects")
        return self._client.request("HEAD", url, **kwargs)

    def options(self, url, **kwargs):
        if "allow_redirects" in kwargs:
            kwargs["follow_redirects"] = kwargs.pop("allow_redirects")
        return self._client.request("OPTIONS", url, **kwargs)


@pytest.fixture
def mock_oauth():
    """Mock OAuth client for OIDC authentication."""
    oauth_mock = MagicMock()
    oidc_mock = MagicMock()
    # Use AsyncMock for async methods so tests can assert calls and awaited behavior
    oidc_mock.authorize_redirect = AsyncMock(return_value=MagicMock(status_code=302, headers={"Location": "https://provider.com/auth"}))
    oidc_mock.authorize_access_token = AsyncMock(
        return_value={
            "access_token": "mock_access_token",
            "id_token": "mock_id_token",
            "userinfo": {
                "email": "test@example.com",
                "name": "Test User",
                "groups": ["test-group"],
            },
        }
    )
    oidc_mock.server_metadata = {"end_session_endpoint": "https://provider.com/logout"}

    oauth_mock.oidc = oidc_mock
    return oauth_mock


@pytest.fixture
def mock_user_management(monkeypatch):
    """Mock user management functions used by OIDC callback processing."""
    mocks = {
        "create_user": MagicMock(),
        "populate_groups": MagicMock(),
        "update_user": MagicMock(),
    }

    # Patch the mlflow_oidc_auth.user functions to use these mocks
    monkeypatch.setattr("mlflow_oidc_auth.user.create_user", mocks["create_user"])
    monkeypatch.setattr("mlflow_oidc_auth.user.populate_groups", mocks["populate_groups"])
    monkeypatch.setattr("mlflow_oidc_auth.user.update_user", mocks["update_user"])

    return mocks


@pytest.fixture
def mock_config():
    """Mock configuration with test values."""
    config_mock = MagicMock()
    config_mock.OIDC_PROVIDER_DISPLAY_NAME = "Test Provider"
    config_mock.OIDC_REDIRECT_URI = "http://localhost:8000/callback"
    config_mock.OIDC_DISCOVERY_URL = "https://provider.com/.well-known/openid_configuration"
    config_mock.OIDC_GROUP_DETECTION_PLUGIN = None
    config_mock.OIDC_GROUPS_ATTRIBUTE = "groups"
    config_mock.OIDC_ADMIN_GROUP_NAME = ["admin-group"]
    config_mock.OIDC_GROUP_NAME = ["user-group", "test-group"]
    config_mock.OIDC_GEN_AI_GATEWAY_ENABLED = False
    return config_mock


@pytest.fixture
def mock_tracking_store():
    """Mock MLflow tracking store."""
    tracking_store_mock = MagicMock()

    # Mock experiment data
    mock_experiment = MagicMock()
    mock_experiment.experiment_id = "123"
    mock_experiment.name = "Test Experiment"
    mock_experiment.tags = {"env": "test"}

    tracking_store_mock.search_experiments.return_value = [mock_experiment]
    tracking_store_mock.search_registered_models.return_value = []
    return tracking_store_mock


@pytest.fixture
def mock_permissions():
    """Mock permission checking functions."""
    permissions_mock = {
        "can_manage_experiment": MagicMock(return_value=True),
        "can_manage_registered_model": MagicMock(return_value=True),
        "can_manage_scorer": MagicMock(return_value=True),
        # Permission helpers may be called synchronously in some test setup;
        # use MagicMock to provide a regular callable that returns the value.
        "get_username": MagicMock(return_value="test@example.com"),
        "get_is_admin": MagicMock(return_value=False),
        # Async variants for FastAPI dependencies which are awaited
        "get_username_async": AsyncMock(return_value="test@example.com"),
        "get_is_admin_async": AsyncMock(return_value=False),
    }
    return permissions_mock


def _mock_filter_manageable_experiments(username, experiments):
    """Mock filter_manageable_experiments that returns all experiments."""
    return experiments


def _mock_filter_manageable_models(username, models):
    """Mock filter_manageable_models that returns all models."""
    return models


def _mock_filter_manageable_prompts(username, prompts):
    """Mock filter_manageable_prompts that returns all prompts."""
    return prompts


@pytest.fixture(autouse=True)
def _patch_router_stores(mock_store):
    """Autouse fixture to patch router module-level 'store' references to the mock_store.

    Some router modules import the module-level 'store' at import-time. Tests often set
    mock_store.list_users etc. but forget to patch the router's copy. This autouse
    fixture ensures the common router modules see the mock_store during tests.
    """
    patches = [
        patch("mlflow_oidc_auth.store.store", mock_store),
        patch("mlflow_oidc_auth.utils.request_helpers_fastapi.store", mock_store),
        patch("mlflow_oidc_auth.utils.batch_permissions.store", mock_store),
        patch("mlflow_oidc_auth.routers.registered_model_permissions.store", mock_store),
        patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store),
        patch("mlflow_oidc_auth.routers.group_permissions.store", mock_store),
        patch("mlflow_oidc_auth.routers.users.store", mock_store),
        patch("mlflow_oidc_auth.routers.experiment_permissions.store", mock_store),
        patch("mlflow_oidc_auth.routers.prompt_permissions.store", mock_store),
        patch("mlflow_oidc_auth.routers.scorers_permissions.store", mock_store),
        patch("mlflow_oidc_auth.routers.gateway_endpoint_permissions.store", mock_store),
        patch("mlflow_oidc_auth.routers.gateway_secret_permissions.store", mock_store),
        patch(
            "mlflow_oidc_auth.routers.gateway_model_definition_permissions.store",
            mock_store,
        ),
        # Patch filter_manageable_* functions at router level so integration tests work
        patch(
            "mlflow_oidc_auth.routers.experiment_permissions.filter_manageable_experiments",
            _mock_filter_manageable_experiments,
        ),
        patch(
            "mlflow_oidc_auth.routers.registered_model_permissions.filter_manageable_models",
            _mock_filter_manageable_models,
        ),
        patch(
            "mlflow_oidc_auth.routers.prompt_permissions.filter_manageable_prompts",
            _mock_filter_manageable_prompts,
        ),
    ]

    for p in patches:
        try:
            p.start()
        except Exception:
            # ignore any that don't apply in this workspace snapshot
            pass

    yield

    for p in patches:
        try:
            p.stop()
        except Exception:
            pass


@pytest.fixture
def authenticated_session():
    """Mock authenticated session data."""
    return {
        "username": "test@example.com",
        "authenticated": True,
        "oauth_state": "test_state",
    }


@pytest.fixture
def unauthenticated_session():
    """Mock unauthenticated session data."""
    return {}


@pytest.fixture
def admin_session():
    """Mock admin user session data."""
    return {"username": "admin@example.com", "authenticated": True, "is_admin": True}


@pytest.fixture
def test_app(mock_store, mock_oauth, mock_config, mock_tracking_store, mock_permissions):
    """Create a test FastAPI application with all routers."""
    # Build test app using the production factory so mounts/middleware match prod

    # Patch runtime dependencies used by middleware, routers and Flask mount
    # Ensure submodules are importable so patch() can resolve dotted names
    try:
        # Import the middleware module so the package object gets the attribute
        from mlflow_oidc_auth.middleware import auth_middleware  # noqa: F401
    except Exception:
        # Ignore import errors here; patches below will attempt best-effort
        pass

    patches = [
        patch("mlflow_oidc_auth.middleware.auth_middleware.store", mock_store),
        patch("mlflow_oidc_auth.oauth.oauth", mock_oauth),
        patch("mlflow_oidc_auth.config.config", mock_config),
        patch(
            "mlflow.server.handlers._get_tracking_store",
            return_value=mock_tracking_store,
        ),
        # Patch _get_tracking_store at router level since it's imported at module-import time
        patch(
            "mlflow_oidc_auth.routers.experiment_permissions._get_tracking_store",
            return_value=mock_tracking_store,
        ),
        patch(
            "mlflow_oidc_auth.utils.can_manage_experiment",
            mock_permissions["can_manage_experiment"],
        ),
        patch(
            "mlflow_oidc_auth.utils.can_manage_scorer",
            mock_permissions["can_manage_scorer"],
        ),
        patch(
            "mlflow_oidc_auth.utils.can_manage_registered_model",
            mock_permissions["can_manage_registered_model"],
        ),
        # utils.* are used synchronously in some places; leave those as MagicMock
        patch("mlflow_oidc_auth.utils.get_username", mock_permissions["get_username"]),
        patch("mlflow_oidc_auth.utils.get_is_admin", mock_permissions["get_is_admin"]),
        # dependencies.* are awaited by FastAPI; patch them with AsyncMock variants
        patch(
            "mlflow_oidc_auth.dependencies.get_username",
            mock_permissions["get_username_async"],
        ),
        patch(
            "mlflow_oidc_auth.dependencies.get_is_admin",
            mock_permissions["get_is_admin_async"],
        ),
        patch(
            "mlflow_oidc_auth.dependencies.can_manage_experiment",
            _deleg_can_manage_experiment,
        ),
        patch("mlflow_oidc_auth.dependencies.can_manage_scorer", _deleg_can_manage_scorer),
        patch(
            "mlflow_oidc_auth.dependencies.can_manage_registered_model",
            _deleg_can_manage_registered_model,
        ),
        # Patch names imported directly into router modules (they were imported at module-import time)
        patch(
            "mlflow_oidc_auth.routers.experiment_permissions.get_is_admin",
            mock_permissions["get_is_admin"],
        ),
        patch(
            "mlflow_oidc_auth.routers.experiment_permissions.get_username",
            mock_permissions["get_username"],
        ),
        patch(
            "mlflow_oidc_auth.routers.scorers_permissions.check_scorer_manage_permission",
            MagicMock(return_value=None),
        ),
        patch(
            "mlflow_oidc_auth.routers.experiment_permissions.can_manage_experiment",
            mock_permissions["can_manage_experiment"],
        ),
        # Patch the module-level 'store' used by request helper functions
        patch("mlflow_oidc_auth.utils.request_helpers_fastapi.store", mock_store),
        patch("mlflow_oidc_auth.store.store", mock_store),
    ]

    # Start all patches before building the test FastAPI app so middleware/routers pick up mocks
    for p in patches:
        try:
            p.start()
        except Exception:
            # If a particular module or attribute can't be patched in this snapshot,
            # skip it and continue. Tests will mock behavior where necessary.
            continue

    try:
        # Build a local FastAPI app similar to production but avoid mounting the real Flask app
        from fastapi import FastAPI
        from starlette.middleware.sessions import (
            SessionMiddleware as StarletteSessionMiddleware,
        )

        from mlflow_oidc_auth.middleware.auth_middleware import AuthMiddleware
        from mlflow_oidc_auth.routers import get_all_routers

        app = FastAPI()
        app.add_middleware(AuthMiddleware)
        app.add_middleware(StarletteSessionMiddleware, secret_key=mock_config.SECRET_KEY)

        for router in get_all_routers():
            app.include_router(router)

        yield app
    finally:
        for p in patches:
            p.stop()


@pytest.fixture
def client(test_app):
    """Create a test client for the FastAPI application."""
    return TestClientWrapper(TestClient(test_app))


@pytest.fixture
def authenticated_client(test_app, authenticated_session):
    """Create a test client with authenticated user."""
    import base64

    client = TestClient(test_app)
    client.headers["Authorization"] = "Basic " + base64.b64encode(b"user@example.com:password").decode()
    return TestClientWrapper(client)


@pytest.fixture
def admin_client(test_app_admin):
    """Create a test client with admin authentication."""
    import base64

    client = TestClient(test_app_admin)
    client.headers["Authorization"] = "Basic " + base64.b64encode(b"admin@example.com:password").decode()
    return TestClientWrapper(client)


@pytest.fixture
def test_app_admin(mock_store, mock_oauth, mock_config, mock_tracking_store, admin_permissions):
    """Create a test FastAPI application with all routers for admin tests."""

    # Ensure middleware submodule exists on package for patch resolution
    try:
        from mlflow_oidc_auth.middleware import auth_middleware  # noqa: F401
    except Exception:
        pass

    patches = [
        patch("mlflow_oidc_auth.store.store", mock_store),
        patch("mlflow_oidc_auth.middleware.auth_middleware.store", mock_store),
        patch("mlflow_oidc_auth.oauth.oauth", mock_oauth),
        patch("mlflow_oidc_auth.config.config", mock_config),
        patch(
            "mlflow.server.handlers._get_tracking_store",
            return_value=mock_tracking_store,
        ),
        # Patch _get_tracking_store at router level since it's imported at module-import time
        patch(
            "mlflow_oidc_auth.routers.experiment_permissions._get_tracking_store",
            return_value=mock_tracking_store,
        ),
        patch(
            "mlflow_oidc_auth.utils.can_manage_experiment",
            admin_permissions["can_manage_experiment"],
        ),
        patch(
            "mlflow_oidc_auth.utils.can_manage_registered_model",
            admin_permissions["can_manage_registered_model"],
        ),
        patch("mlflow_oidc_auth.utils.can_manage_scorer", MagicMock(return_value=True)),
        # utils.* remain sync mocks
        patch("mlflow_oidc_auth.utils.get_username", admin_permissions["get_username"]),
        patch("mlflow_oidc_auth.utils.get_is_admin", admin_permissions["get_is_admin"]),
        # dependencies.* patched to async variants for FastAPI awaits
        patch(
            "mlflow_oidc_auth.dependencies.get_username",
            admin_permissions["get_username_async"],
        ),
        patch(
            "mlflow_oidc_auth.dependencies.get_is_admin",
            admin_permissions["get_is_admin_async"],
        ),
        # Also patch router-level imported names for admin app
        patch(
            "mlflow_oidc_auth.routers.experiment_permissions.get_is_admin",
            admin_permissions["get_is_admin"],
        ),
        patch(
            "mlflow_oidc_auth.routers.experiment_permissions.get_username",
            admin_permissions["get_username"],
        ),
        patch(
            "mlflow_oidc_auth.routers.experiment_permissions.can_manage_experiment",
            admin_permissions["can_manage_experiment"],
        ),
        patch(
            "mlflow_oidc_auth.dependencies.can_manage_experiment",
            _deleg_can_manage_experiment,
        ),
        patch(
            "mlflow_oidc_auth.dependencies.can_manage_registered_model",
            _deleg_can_manage_registered_model,
        ),
        patch("mlflow_oidc_auth.dependencies.can_manage_scorer", _deleg_can_manage_scorer),
        # Patch request helper module-level store
        patch("mlflow_oidc_auth.utils.request_helpers_fastapi.store", mock_store),
        patch(
            "mlflow_oidc_auth.routers.prompt_permissions.check_admin_permission",
            MagicMock(return_value="admin@example.com"),
        ),
        patch(
            "mlflow_oidc_auth.routers.prompt_permissions.get_username",
            admin_permissions["get_username"],
        ),
        patch(
            "mlflow_oidc_auth.routers.prompt_permissions.get_is_admin",
            admin_permissions["get_is_admin"],
        ),
        patch(
            "mlflow_oidc_auth.routers.scorers_permissions.check_scorer_manage_permission",
            MagicMock(return_value=None),
        ),
    ]

    for p in patches:
        try:
            p.start()
        except Exception:
            continue

    try:
        from fastapi import FastAPI
        from starlette.middleware.sessions import (
            SessionMiddleware as StarletteSessionMiddleware,
        )

        from mlflow_oidc_auth.middleware.auth_middleware import AuthMiddleware
        from mlflow_oidc_auth.routers import get_all_routers

        app = FastAPI()
        app.add_middleware(AuthMiddleware)
        app.add_middleware(StarletteSessionMiddleware, secret_key=mock_config.SECRET_KEY)

        for router in get_all_routers():
            app.include_router(router)

        yield app
    finally:
        for p in patches:
            p.stop()


@pytest.fixture
def mock_request_with_session():
    """Create a mock FastAPI request with session."""

    def _create_request(session_data: Optional[Dict[str, Any]] = None):
        request_mock = MagicMock()
        request_mock.session = session_data or {}
        request_mock.base_url = "http://localhost:8000"
        request_mock.query_params = {}
        return request_mock

    return _create_request


@pytest.fixture
def sample_experiment_permissions():
    """Sample experiment permission data for testing."""
    return [
        ExperimentPermissionEntity(experiment_id="123", permission=Permission.MANAGE.name, user_id=1),
        ExperimentPermissionEntity(experiment_id="456", permission=Permission.READ.name, user_id=1),
    ]


@pytest.fixture
def sample_users_data():
    """Sample user data for testing."""
    return [
        {
            "username": "admin@example.com",
            "display_name": "Admin User",
            "is_admin": True,
            "is_service_account": False,
        },
        {
            "username": "user@example.com",
            "display_name": "Regular User",
            "is_admin": False,
            "is_service_account": False,
        },
        {
            "username": "service@example.com",
            "display_name": "Service Account",
            "is_admin": False,
            "is_service_account": True,
        },
    ]


@pytest.fixture
def mock_logger():
    """Mock logger for testing."""
    with patch("mlflow_oidc_auth.logger.get_logger") as mock_get_logger:
        logger_mock = MagicMock()
        mock_get_logger.return_value = logger_mock
        yield logger_mock


@pytest.fixture
def admin_permissions():
    """Mock permission checking functions for admin user."""
    permissions_mock = {
        "can_manage_experiment": MagicMock(return_value=True),
        "can_manage_registered_model": MagicMock(return_value=True),
        # Admin permission helpers used in fixture wiring are sync-callable
        "get_username": MagicMock(return_value="admin@example.com"),
        "get_is_admin": MagicMock(return_value=True),
        # Async variants for dependencies
        "get_username_async": AsyncMock(return_value="admin@example.com"),
        "get_is_admin_async": AsyncMock(return_value=True),
    }
    return permissions_mock



from flask import Response, g, request
from mlflow.protos.model_registry_pb2 import CreateRegisteredModel, DeleteRegisteredModel, RenameRegisteredModel, SearchRegisteredModels
from mlflow.protos.service_pb2 import (
    CreateExperiment,
    CreateGatewayEndpoint,
    CreateGatewayModelDefinition,
    CreateGatewaySecret,
    DeleteGatewayEndpoint,
    DeleteGatewayModelDefinition,
    DeleteGatewaySecret,
    DeleteScorer,
    ListGatewayEndpoints,
    ListGatewayModelDefinitions,
    ListGatewaySecretInfos,
    RegisterScorer,
    SearchExperiments,
    SearchLoggedModels,
    UpdateGatewayEndpoint,
)
from mlflow.server.handlers import _get_model_registry_store, _get_request_message, _get_tracking_store, catch_mlflow_exception, get_endpoints
from mlflow.utils.proto_json_utils import message_to_json, parse_dict
from mlflow.utils.search_utils import SearchUtils

from mlflow_oidc_auth.bridge import get_fastapi_admin_status, get_fastapi_username
from mlflow_oidc_auth.logger import get_logger
from mlflow_oidc_auth.permissions import MANAGE
from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.utils import can_read_experiment, can_read_registered_model, get_model_name
from mlflow_oidc_auth.utils.permissions import can_read_gateway_endpoint, can_read_gateway_model_definition, can_read_gateway_secret


def _set_can_manage_experiment_permission(resp: Response):
    response_message = CreateExperiment.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    experiment_id = response_message.experiment_id
    username = get_fastapi_username()
    store.create_experiment_permission(experiment_id, username, MANAGE.name)


def _set_can_manage_registered_model_permission(resp: Response):
    response_message = CreateRegisteredModel.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    name = response_message.registered_model.name
    username = get_fastapi_username()
    store.create_registered_model_permission(name, username, MANAGE.name)


def _delete_can_manage_registered_model_permission(resp: Response):
    """
    Delete registered model permission when the model is deleted.

    We need to do this because the primary key of the registered model is the name,
    unlike the experiment where the primary key is experiment_id (UUID). Therefore,
    we have to delete the permission record when the model is deleted otherwise it
    conflicts with the new model registered with the same name.
    """
    # Get model name from request context because it's not available in the response
    model_name = get_model_name()
    if not model_name:
        return
    store.wipe_group_model_permissions(model_name)
    store.wipe_registered_model_permissions(model_name)


def _get_after_request_handler(request_class):
    return AFTER_REQUEST_PATH_HANDLERS.get(request_class)


def _filter_search_experiments(resp: Response):
    if get_fastapi_admin_status():
        return

    response_message = SearchExperiments.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    request_message = _get_request_message(SearchExperiments())

    username = get_fastapi_username()

    # Filter out unreadable experiments from the current response page.
    for e in list(response_message.experiments):
        if not can_read_experiment(e.experiment_id, username):
            response_message.experiments.remove(e)

    # Re-fetch to fill max_results, preserving MLflow pagination semantics.
    tracking_store = _get_tracking_store()
    while len(response_message.experiments) < request_message.max_results and response_message.next_page_token != "":
        refetched = tracking_store.search_experiments(
            view_type=request_message.view_type,
            max_results=request_message.max_results,
            order_by=request_message.order_by,
            filter_string=request_message.filter,
            page_token=response_message.next_page_token,
        )

        remaining = request_message.max_results - len(response_message.experiments)
        refetched = refetched[:remaining]
        if len(refetched) == 0:
            response_message.next_page_token = ""
            break

        readable_proto = [e.to_proto() for e in refetched if can_read_experiment(e.experiment_id, username)]
        response_message.experiments.extend(readable_proto)

        start_offset = SearchUtils.parse_start_offset_from_page_token(response_message.next_page_token)
        final_offset = start_offset + len(refetched)
        response_message.next_page_token = SearchUtils.create_page_token(final_offset)

    resp.data = message_to_json(response_message)


def _filter_search_registered_models(resp: Response):
    if get_fastapi_admin_status():
        return

    response_message = SearchRegisteredModels.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    request_message = _get_request_message(SearchRegisteredModels())

    username = get_fastapi_username()

    # Filter out unreadable models from the current response page.
    for rm in list(response_message.registered_models):
        if not can_read_registered_model(rm.name, username):
            response_message.registered_models.remove(rm)

    # Re-fetch to fill max_results, preserving MLflow pagination semantics.
    model_registry_store = _get_model_registry_store()
    while len(response_message.registered_models) < request_message.max_results and response_message.next_page_token != "":
        refetched = model_registry_store.search_registered_models(
            filter_string=request_message.filter,
            max_results=request_message.max_results,
            order_by=request_message.order_by,
            page_token=response_message.next_page_token,
        )
        remaining = request_message.max_results - len(response_message.registered_models)
        refetched = refetched[:remaining]
        if len(refetched) == 0:
            response_message.next_page_token = ""
            break

        readable_proto = [rm.to_proto() for rm in refetched if can_read_registered_model(rm.name, username)]
        response_message.registered_models.extend(readable_proto)

        start_offset = SearchUtils.parse_start_offset_from_page_token(response_message.next_page_token)
        final_offset = start_offset + len(refetched)
        response_message.next_page_token = SearchUtils.create_page_token(final_offset)

    resp.data = message_to_json(response_message)


def _filter_search_logged_models(resp: Response) -> None:
    """
    Filter out unreadable logged models from the search results.
    """
    if get_fastapi_admin_status():
        return

    response_message = SearchLoggedModels.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    request_message = _get_request_message(SearchLoggedModels())

    username = get_fastapi_username()

    # Remove unreadable models from the current response page.
    for m in list(response_message.models):
        if not can_read_experiment(m.info.experiment_id, username):
            response_message.models.remove(m)

    from mlflow.utils.search_utils import SearchLoggedModelsPaginationToken as Token

    max_results = request_message.max_results
    params = {
        "experiment_ids": list(request_message.experiment_ids),
        "filter_string": request_message.filter or None,
        "order_by": (
            [
                {
                    "field_name": ob.field_name,
                    "ascending": ob.ascending,
                    "dataset_name": ob.dataset_name,
                    "dataset_digest": ob.dataset_digest,
                }
                for ob in request_message.order_by
            ]
            if request_message.order_by
            else None
        ),
    }

    next_page_token = response_message.next_page_token or None
    tracking_store = _get_tracking_store()

    while len(response_message.models) < max_results and next_page_token is not None:
        batch = tracking_store.search_logged_models(max_results=max_results, page_token=next_page_token, **params)
        is_last_page = batch.token is None
        offset = Token.decode(next_page_token).offset if next_page_token else 0
        last_index = len(batch) - 1

        for index, model in enumerate(batch):
            if not can_read_experiment(model.experiment_id, username):
                continue

            response_message.models.append(model.to_proto())
            if len(response_message.models) >= max_results:
                next_page_token = None if is_last_page and index == last_index else Token(offset=offset + index + 1, **params).encode()
                break
        else:
            next_page_token = None if is_last_page else Token(offset=offset + max_results, **params).encode()

    response_message.next_page_token = next_page_token or ""
    resp.data = message_to_json(response_message)


def _rename_registered_model_permission(resp: Response):
    """
    A model registry can be assigned to multiple users or groups with different permissions.
    Changing the model registry name must be propagated to all users or groups.
    """
    data = request.get_json(force=True, silent=True)
    name = data.get("name") if data else None
    new_name = data.get("new_name") if data else None
    if not name or not new_name:
        # Defensive no-op: avoid turning a successful rename into a 500 in after_request.
        return
    store.rename_registered_model_permissions(name, new_name)
    store.rename_group_model_permissions(name, new_name)


def _set_can_manage_scorer_permission(resp: Response):
    """Create MANAGE scorer permission for the scorer creator."""

    response_message = RegisterScorer.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    experiment_id = response_message.experiment_id
    scorer_name = response_message.name
    username = get_fastapi_username()
    store.create_scorer_permission(experiment_id, scorer_name, username, MANAGE.name)


def _delete_scorer_permissions_cascade(resp: Response):
    """Delete all scorer permissions when a scorer is deleted."""

    data = request.get_json(force=True, silent=True) or {}
    experiment_id = data.get("experiment_id")
    scorer_name = data.get("name")
    if experiment_id and scorer_name:
        store.delete_scorer_permissions_for_scorer(str(experiment_id), str(scorer_name))


def _set_can_manage_gateway_endpoint_permission(resp: Response):
    """Create MANAGE gateway endpoint permission for the endpoint creator."""
    response_message = CreateGatewayEndpoint.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    name = response_message.endpoint.name
    username = get_fastapi_username()
    store.create_gateway_endpoint_permission(name, username, MANAGE.name)


def _rename_gateway_endpoint_permission(resp: Response):
    """Propagate endpoint name changes to all user and group permission records.

    The old name was stashed in ``flask.g`` by the before-request validator.
    The new name is read from the response.
    """
    old_name = getattr(g, "_updating_gateway_endpoint_old_name", None)
    if not old_name:
        return

    response_message = UpdateGatewayEndpoint.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    new_name = response_message.endpoint.name

    if not new_name or new_name == old_name:
        # Name unchanged — nothing to do.
        return

    try:
        store.rename_gateway_endpoint_permissions(old_name, new_name)
    except Exception:
        get_logger().warning(f"Failed to rename gateway endpoint permissions from '{old_name}' to '{new_name}'")


def _set_can_manage_gateway_secret_permission(resp: Response):
    """Create MANAGE gateway secret permission for the secret creator."""
    response_message = CreateGatewaySecret.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    secret_name = response_message.secret.secret_name
    username = get_fastapi_username()
    store.create_gateway_secret_permission(secret_name, username, MANAGE.name)


def _set_can_manage_gateway_model_definition_permission(resp: Response):
    """Create MANAGE gateway model definition permission for the model definition creator."""
    response_message = CreateGatewayModelDefinition.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    name = response_message.model_definition.name
    username = get_fastapi_username()
    store.create_gateway_model_definition_permission(name, username, MANAGE.name)


def _filter_list_gateway_endpoints(resp: Response) -> None:
    """Filter out gateway endpoints the user cannot read."""
    if get_fastapi_admin_status():
        return

    response_message = ListGatewayEndpoints.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    username = get_fastapi_username()
    logger = get_logger()

    for endpoint in list(response_message.endpoints):
        if not can_read_gateway_endpoint(endpoint.name, username):
            response_message.endpoints.remove(endpoint)
            logger.debug(f"Filtered gateway endpoint '{endpoint.name}' for user '{username}'")

    resp.data = message_to_json(response_message)


def _filter_list_gateway_secrets(resp: Response) -> None:
    """Filter out gateway secrets the user cannot read."""
    if get_fastapi_admin_status():
        return

    response_message = ListGatewaySecretInfos.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    username = get_fastapi_username()
    logger = get_logger()

    for secret in list(response_message.secrets):
        if not can_read_gateway_secret(secret.secret_name, username):
            response_message.secrets.remove(secret)
            logger.debug(f"Filtered gateway secret '{secret.secret_name}' for user '{username}'")

    resp.data = message_to_json(response_message)


def _filter_list_gateway_model_definitions(resp: Response) -> None:
    """Filter out gateway model definitions the user cannot read."""
    if get_fastapi_admin_status():
        return

    response_message = ListGatewayModelDefinitions.Response()  # type: ignore
    parse_dict(resp.json, response_message)
    username = get_fastapi_username()
    logger = get_logger()

    for model_def in list(response_message.model_definitions):
        if not can_read_gateway_model_definition(model_def.name, username):
            response_message.model_definitions.remove(model_def)
            logger.debug(f"Filtered gateway model definition '{model_def.name}' for user '{username}'")

    resp.data = message_to_json(response_message)


def _delete_gateway_endpoint_permissions_cascade(resp: Response) -> None:
    """Delete all permissions for a gateway endpoint after it is deleted."""
    name = getattr(g, "_deleting_gateway_endpoint_name", None)
    if not name:
        return
    try:
        store.wipe_gateway_endpoint_permissions(name)
    except Exception:
        get_logger().warning(f"Failed to cascade-delete permissions for gateway endpoint '{name}'")


def _delete_gateway_secret_permissions_cascade(resp: Response) -> None:
    """Delete all permissions for a gateway secret after it is deleted."""
    name = getattr(g, "_deleting_gateway_secret_name", None)
    if not name:
        return
    try:
        store.wipe_gateway_secret_permissions(name)
    except Exception:
        get_logger().warning(f"Failed to cascade-delete permissions for gateway secret '{name}'")


def _delete_gateway_model_definition_permissions_cascade(resp: Response) -> None:
    """Delete all permissions for a gateway model definition after it is deleted."""
    name = getattr(g, "_deleting_gateway_model_definition_name", None)
    if not name:
        return
    try:
        store.wipe_gateway_model_definition_permissions(name)
    except Exception:
        get_logger().warning(f"Failed to cascade-delete permissions for gateway model definition '{name}'")


AFTER_REQUEST_PATH_HANDLERS = {
    CreateExperiment: _set_can_manage_experiment_permission,
    CreateRegisteredModel: _set_can_manage_registered_model_permission,
    DeleteRegisteredModel: _delete_can_manage_registered_model_permission,
    SearchExperiments: _filter_search_experiments,
    SearchLoggedModels: _filter_search_logged_models,
    SearchRegisteredModels: _filter_search_registered_models,
    RenameRegisteredModel: _rename_registered_model_permission,
    RegisterScorer: _set_can_manage_scorer_permission,
    DeleteScorer: _delete_scorer_permissions_cascade,
    CreateGatewayEndpoint: _set_can_manage_gateway_endpoint_permission,
    CreateGatewaySecret: _set_can_manage_gateway_secret_permission,
    CreateGatewayModelDefinition: _set_can_manage_gateway_model_definition_permission,
    UpdateGatewayEndpoint: _rename_gateway_endpoint_permission,
    DeleteGatewayEndpoint: _delete_gateway_endpoint_permissions_cascade,
    DeleteGatewaySecret: _delete_gateway_secret_permissions_cascade,
    DeleteGatewayModelDefinition: _delete_gateway_model_definition_permissions_cascade,
    ListGatewayEndpoints: _filter_list_gateway_endpoints,
    ListGatewaySecretInfos: _filter_list_gateway_secrets,
    ListGatewayModelDefinitions: _filter_list_gateway_model_definitions,
}

_our_handlers = set(AFTER_REQUEST_PATH_HANDLERS.values())
AFTER_REQUEST_HANDLERS = {
    (http_path, method): handler
    for http_path, handler, methods in get_endpoints(_get_after_request_handler)
    for method in methods
    if handler in _our_handlers and "/graphql" not in http_path
}


@catch_mlflow_exception
def after_request_hook(resp: Response):
    if 400 <= resp.status_code < 600:
        return resp

    if handler := AFTER_REQUEST_HANDLERS.get((request.path, request.method)):
        handler(resp)
    return resp



from dataclasses import dataclass
from typing import Optional, Any, Dict

from ._base import PermissionBase, RegexPermissionBase


@dataclass
class GatewayEndpointPermission(PermissionBase):
    def __init__(self, endpoint_id: str, permission: str, user_id: Optional[int] = None, group_id: Optional[int] = None):
        super().__init__(instance=endpoint_id, permission=permission, user_id=user_id, group_id=group_id)

    @property
    def endpoint_id(self) -> str:
        return self.instance

    def to_json(self) -> Dict[str, Any]:
        d = super().to_json()
        d["endpoint_id"] = d.pop("instance")
        return d

    @classmethod
    def from_json(cls, d: Dict[str, Any]) -> "GatewayEndpointPermission":
        user_id = d.get("user_id")
        group_id = d.get("group_id")
        if user_id is not None:
            try:
                user_id = int(user_id)
            except (TypeError, ValueError):
                raise ValueError("user_id must be an integer")
        if group_id is not None:
            try:
                group_id = int(group_id)
            except (TypeError, ValueError):
                raise ValueError("group_id must be an integer")
        return cls(
            endpoint_id=d["endpoint_id"],
            permission=d["permission"],
            user_id=user_id,
            group_id=group_id,
        )


@dataclass(init=False)
class GatewayEndpointRegexPermission(RegexPermissionBase):
    def __init__(self, id_, regex, priority, user_id, permission):
        super().__init__(
            id=id_,
            regex=regex,
            priority=priority,
            permission=permission,
            user_id=user_id,
        )

    @classmethod
    def from_json(cls, d: Dict[str, Any]) -> "GatewayEndpointRegexPermission":
        user_id = d.get("user_id")
        if user_id is not None:
            try:
                user_id = int(user_id)
            except (TypeError, ValueError):
                raise ValueError("user_id must be an integer")
        return cls(
            id_=d["id"],
            regex=d["regex"],
            priority=d["priority"],
            user_id=user_id,
            permission=d.get("permission"),
        )


@dataclass(init=False)
class GatewayEndpointGroupRegexPermission(RegexPermissionBase):
    def __init__(self, id_, regex, priority, group_id, permission):
        super().__init__(
            id=id_,
            regex=regex,
            priority=priority,
            permission=permission,
            group_id=group_id,
        )

    @classmethod
    def from_json(cls, d: Dict[str, Any]) -> "GatewayEndpointGroupRegexPermission":
        group_id = d.get("group_id")
        if group_id is not None:
            try:
                group_id = int(group_id)
            except (TypeError, ValueError):
                raise ValueError("group_id must be an integer")
        return cls(
            id_=d["id"],
            regex=d["regex"],
            priority=d["priority"],
            group_id=group_id,
            permission=d.get("permission"),
        )
